# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.


from bs4 import BeautifulSoup
import requests

def bandcamp_search(artist, callback=None):

    try:
        page = requests.get("https://bandcamp.com/search?q=" + artist)
        html = BeautifulSoup(page.text, 'html.parser')
        results = html.find_all("div", {"class": "result-info"})
        for result in results:
            children = result.findChildren("div")
            okay = False
            for child in children:
                if child.string and "ARTIST" in child.string:
                    okay = True
                    break
            if not okay:
                continue
            for child in children:
                if child["class"][0] == "heading":
                    if child.a.string.strip().lower() == artist.lower():
                        url = child.a["href"].split("?")[0]
                        if callback:
                            callback(url)
                        return url

    except:
        print("Bandcamp search error")

    if callback:
        callback(None)
    return None




# Tauon Music Box - Web interface module

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.
import fcntl
import html
import time
import random
import json
import io
import os
import subprocess
from http.server import HTTPServer
from http.server import BaseHTTPRequestHandler
from socketserver import ThreadingMixIn
from t_modules.t_extra import Timer

def webserve(pctl, prefs, gui, album_art_gen, install_directory, strings, tauon):

    if prefs.enable_web is False:
        return 0

    def get_broadcast_track():
        if pctl.broadcast_active is False:
            return None, None
        delay = 6
        tr = None
        t = time.time()
        for item in reversed(pctl.broadcast_update_train):
            if 20 > t - item[2] > delay:
                tr = item
                break
        if tr is None:
            return None, None
        else:
            return tr[0], tr[1]

    chunker = tauon.chunker
    gui.web_running = True

    class Server(BaseHTTPRequestHandler):

        def send_file(self, path, mime):
            self.send_response(200)
            self.send_header("Content-type", mime)
            self.end_headers()

            with open(path, "rb") as f:
                self.wfile.write(f.read())

        def do_GET(self):

            path = self.path

            # print(self.headers)

            if path == "/radio/":
                self.send_response(302)
                self.send_header('Location', "/radio")
                self.end_headers()

            elif path == "/radio":
                self.send_file(install_directory + "/templates/radio.html", "text/html")
            elif path == "/favicon.ico":
                self.send_file(install_directory + "/assets/favicon.ico", 'image/x-icon')
            elif path == "/radio/radio.js":
                self.send_file(install_directory + "/templates/radio.js", "application/javascript")
            elif path == "/radio/theme.css":
                self.send_file(install_directory + "/templates/theme.css", "text/css")
            elif path == "/radio/logo-bg.png":
                self.send_file(install_directory + "/templates/logo-bg.png", 'image/png')

            elif path == "/radio/update_radio":
                self.send_response(200)
                self.send_header("Content-type", "application/json")

                track_id, p = get_broadcast_track()
                if track_id is not None:

                    track = pctl.master_library[track_id]
                    if track.length > 2:
                        position = p / track.length
                    else:
                        position = 0
                    data = {"position": position,
                            "index": track.index,
                            "port": str(prefs.broadcast_port)}

                    data = json.dumps(data).replace(" ", "").encode()
                    self.send_header("Content-length", str(len(data)))
                    self.end_headers()
                    self.wfile.write(data)

                else:
                    data = {"position": 0,
                            "index": -1}
                    data = json.dumps(data).replace(" ", "").encode()
                    self.send_header("Content-length", str(len(data)))
                    self.end_headers()
                    self.wfile.write(data)

            elif path == "/radio/getpic":
                self.send_response(200)
                self.send_header("Content-type", "application/json")


                track_id, p = get_broadcast_track()

                if track_id is not None:

                    track = pctl.master_library[track_id]

                    # Lyrics ---
                    lyrics = ""

                    if prefs.radio_page_lyrics:
                        lyrics = tauon.synced_to_static_lyrics.get(track)
                        lyrics = html.escape(lyrics).replace("\r\n", "\n").replace("\r", "\n").replace("\n", "<br>")
                    try:
                        base64 = album_art_gen.get_base64(track, (300, 300)).decode()

                        data = {
                            "index": track_id,
                            "image": base64,
                            "title": track.title,
                            "artist": track.artist,
                            "album": track.album,
                            "lyrics": lyrics}

                        data = json.dumps(data).encode()
                        self.send_header("Content-length", str(len(data)))
                        self.end_headers()
                        self.wfile.write(data)
                    except:
                        # Failed getting image
                        data = {
                            "index": track_id,
                            "image": "None",
                            "title": track.title,
                            "artist": track.artist,
                            "album": track.album,
                            "lyrics": lyrics}

                        data = json.dumps(data).encode()
                        self.send_header("Content-length", str(len(data)))
                        self.end_headers()
                        self.wfile.write(data)
                else:
                    # Broadcast is not active
                    data = {
                        "index": -1,
                        "image": "None",
                        "title": "",
                        "artist": "- - Broadcast Offline - -",
                        "album": "",
                        "lyrics": ""}

                    data = json.dumps(data).encode()
                    self.send_header("Content-length", str(len(data)))
                    self.end_headers()
                    self.wfile.write(data)

            elif path == "/stream.ogg":

                ip = self.client_address[0]

                self.send_response(200)
                self.send_header("Content-type", "audio/ogg")
                #self.send_header("Transfer-Encoding", "chunked")
                self.end_headers()

                position = max(chunker.master_count - 7, 1)
                id = random.random()

                for header in chunker.headers:
                    #self.wfile.write(hex(len(header))[2:].encode())
                    #self.wfile.write("\r\n".encode())
                    self.wfile.write(header)
                    #self.wfile.write("\r\n".encode())
                while True:
                    if not pctl.broadcast_active:
                        return
                    if 1 < position < chunker.master_count:
                        while 1 < position < chunker.master_count:
                            if not pctl.broadcast_active:
                                return
                            #self.wfile.write(hex(len(chunker.chunks[position]))[2:].encode())

                            #self.wfile.write("\r\n".encode())
                            self.wfile.write(chunker.chunks[position])
                            #self.wfile.write("\r\n".encode())

                            position += 1
                    else:
                        time.sleep(0.01)
                        chunker.clients[id] = (ip, time.time())
            else:
                self.send_response(404)
                self.end_headers()
                self.wfile.write(b"404 Not found")

    class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
        pass

    try:
        httpd = ThreadedHTTPServer(("0.0.0.0", prefs.metadata_page_port), Server)
        tauon.radio_server = httpd
        httpd.serve_forever()
        httpd.server_close()
    except OSError:
        print("Not starting radio page server, already running?")


def webserve2(pctl, prefs, gui, album_art_gen, install_directory, strings, tauon):

    play_timer = Timer()

    class Server(BaseHTTPRequestHandler):

        def run_command(self, callback):
            self.send_response(200)
            #self.send_header("Content-type", "application/json")
            self.end_headers()
            callback()
            self.wfile.write(b"OK")

        def parse_trail(self, text):

            params = {}
            both = text.split("?")
            levels = both[0].split("/")
            if len(both) > 1:
                pairs = both[2].split("&")
                for p in pairs:
                    aa, bb = p.split("=")
                    params[aa] = bb

            return levels, params

        def get_track(self, track_position, playlist_index=None, track=None, album_id=-1):
            if track is None:
                if playlist_index is None:
                    playlist = pctl.multi_playlist[pctl.active_playlist_playing][2]
                else:
                    playlist = pctl.multi_playlist[playlist_index][2]
                track_id = playlist[track_position]
                track = pctl.g(track_id)

            data = {}
            data["title"] = track.title
            data["artist"] = track.artist
            data["album"] = track.album
            data["album_artist"] = track.album_artist
            if not track.album_artist:
                data["album_artist"] = track.artist
            data["duration"] = int(track.length * 1000)
            data["id"] = track.index
            data["position"] = track_position
            data["album_id"] = album_id
            data["has_lyrics"] = track.lyrics != ""
            data["track_number"] = str(track.track_number).lstrip("0")
            data["can_download"] = not track.is_cue and not track.is_network

            return data

        def send_file(self, path, mime):

            range_req = False
            start = 0
            end = 0

            if "Range" in self.headers:
                range_req = True
                b = self.headers["Range"].split("=")[1]
                start, end = b.split("-")
                start = int(start)

            with open(path, "rb") as f:

                f.seek(0, 2)
                length = f.tell()
                f.seek(0, 0)

                l = str(length)

                remain = length - start

                if range_req:
                    self.send_response(206)
                    self.send_header("Content-type", mime)
                    self.send_header("Content-Range", f"bytes={start}-/{l}")
                    self.send_header("Content-Length", str(remain))
                    f.seek(start)

                else:
                    self.send_response(200)
                    self.send_header("Content-type", mime)
                    self.send_header("Content-Length", l)

                self.end_headers()

                while True:
                    data = f.read(5000)
                    if not data:
                        break
                    self.wfile.write(data)

        def do_GET(self):

            path = self.path

            if path.startswith("/api1/pic/small/"):
                value = path[16:]
                if value.isalnum() and int(value) in pctl.master_library:
                    track = pctl.g(int(value))
                    raw = album_art_gen.save_thumb(track, (75, 75), "")
                    if raw:
                        self.send_response(200)
                        self.send_header("Content-type", "image/jpg")
                        self.end_headers()
                        self.wfile.write(raw.read())
                    else:
                        self.send_response(404)
                        self.end_headers()
                        self.wfile.write(b"No image found")

                else:
                    self.send_response(404)
                    self.end_headers()
                    self.wfile.write(b"Invalid parameter")

            if path.startswith("/api1/pic/medium/"):
                value = path[17:]
                if value.isalnum() and int(value) in pctl.master_library:
                    track = pctl.g(int(value))
                    raw = album_art_gen.save_thumb(track, (1000, 1000), "")
                    if raw:
                        self.send_response(200)
                        self.send_header("Content-type", "image/jpg")
                        self.end_headers()
                        self.wfile.write(raw.read())
                    else:
                        self.send_response(404)
                        self.end_headers()
                        self.wfile.write(b"No image found")

                else:
                    self.send_response(404)
                    self.end_headers()
                    self.wfile.write(b"Invalid parameter")

            if path.startswith("/api1/lyrics/"):
                value = path[13:]
                if value.isalnum() and int(value) in pctl.master_library:
                    track = pctl.g(int(value))
                    data = {}
                    data["track_id"] = track.index
                    data["lyrics_text"] = track.lyrics

                    self.send_response(200)
                    self.send_header("Content-type", "application/json")
                    self.end_headers()
                    data = json.dumps(data).encode()
                    self.wfile.write(data)
                else:
                    self.send_response(404)
                    self.end_headers()
                    self.wfile.write(b"Invalid parameter")

            # elif path.startswith("/api1/stream/"):
            #     param = path[13:]
            #
            #     if param.isdigit() and int(param) in pctl.master_library:
            #         track = pctl.master_library[int(param)]
            #         mime = "audio/mpeg"
            #         #mime = "audio/ogg"
            #         self.send_response(200)
            #         self.send_header("Content-type", mime)
            #         self.end_headers()
            #
            #         cmd = ["ffmpeg", "-i", track.fullpath, "-c:a", "libopus", "-f", "ogg", "-"]
            #         #cmd = ["ffmpeg", "-i", track.fullpath, "-c:a", "libvorbis", "-f", "ogg", "-"]
            #         #cmd = ["ffmpeg", "-i", track.fullpath, "-c:a", "libmp3lame", "-f", "mp3", "-"]
            #         encoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            #         while True:
            #             data = encoder.stdout.read(1024)
            #             if data:
            #                 self.wfile.write(data)
            elif path.startswith("/api1/playinghit/"):
                param = path[17:]
                if param.isdigit() and int(param) in pctl.master_library:
                    t = play_timer.hit()
                    if 0 < t < 5:
                        tauon.star_store.add(int(param), t)

                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                self.wfile.write(b"OK")

            elif path.startswith("/api1/file/"):
                param = path[11:]

                #print(self.headers)
                play_timer.hit()

                if param.isdigit() and int(param) in pctl.master_library:
                    track = pctl.master_library[int(param)]
                    mime = "audio/mpeg"
                    if track.file_ext == "FLAC":
                        mime = "audio/flac"
                    if track.file_ext == "OGG" or track.file_ext == "OPUS" or track.file_ext == "OGA":
                        mime = "audio/ogg"
                    if track.file_ext == "M4A":
                        mime = "audio/mp4"
                    self.send_file(track.fullpath, mime)

            elif path.startswith("/api1/start/"):

                levels, _ = self.parse_trail(path)
                if len(levels) == 5:
                    playlist = levels[3]
                    position = levels[4]
                    if playlist.isdigit() and position.isdigit():
                        position = int(position)
                        playlist = int(playlist)
                        pl = tauon.id_to_pl(int(playlist))
                        if pl is not None and pl < len(pctl.multi_playlist):
                            playlist = pctl.multi_playlist[pl][2]
                            if position < len(playlist):
                                tauon.switch_playlist(pl, cycle=False, quiet=True)
                                pctl.jump(playlist[position], position)

                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                self.wfile.write(b"OK")

            elif path == "/api1/play":
                self.run_command(tauon.pctl.play)
            elif path == "/api1/pause":
                self.run_command(tauon.pctl.pause_only)
            elif path == "/api1/next":
                self.run_command(tauon.pctl.advance)
            elif path == "/api1/back":
                self.run_command(tauon.pctl.back)
            elif path == "/api1/shuffle":
                self.run_command(tauon.toggle_random)
            elif path == "/api1/repeat":
                self.run_command(tauon.toggle_repeat)
            elif path == "/api1/version":
                data = {"version": 1}
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                data = json.dumps(data).encode()
                self.wfile.write(data)

            elif path == "/api1/playlists":

                l = []
                for item in pctl.multi_playlist:
                    p = {}
                    p["name"] = item[0]
                    p["id"] = str(item[6])
                    p["count"] = len(item[2])
                    l.append(p)
                data = {"playlists": l}
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                data = json.dumps(data).encode()
                self.wfile.write(data)

            elif path.startswith("/api1/albumtracks/"):
                # Get tracks that appear in an album /albumtracks/plid/albumid
                levels, _ = self.parse_trail(path)
                l = []
                if len(levels) == 5 and levels[3].isdigit() and levels[4].isdigit():
                    pl = tauon.id_to_pl(int(levels[3]))
                    if pl is not None:
                        _, album, _ = tauon.get_album_info(int(levels[4]), pl)
                        # print(album)
                        for p in album:
                            l.append(self.get_track(p, pl, album_id=int(levels[4])))

                data = {"tracks": l}
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                data = json.dumps(data).encode()
                self.wfile.write(data)

            elif path.startswith("/api1/trackposition/"):
                # get track /trackposition/plid/playlistposition
                levels, _ = self.parse_trail(path)
                if len(levels) == 5 and levels[3].isdigit() and levels[4].isdigit():
                    pl = tauon.id_to_pl(int(levels[3]))
                    if pl is not None:

                        data = self.get_track(int(levels[4]), pl)

                        playlist = pctl.multi_playlist[pl][2]
                        p = int(levels[4])
                        if p < len(playlist):
                            track = pctl.g(playlist[p])
                            while True:
                                if p < 0 or pctl.g(playlist[p]).parent_folder_path != track.parent_folder_path:
                                    p += 1
                                    break
                                p -= 1
                            data["album_id"] = p

                            self.send_response(200)
                            self.send_header("Content-type", "application/json")
                            self.end_headers()
                            data = json.dumps(data).encode()
                            self.wfile.write(data)
                        else:
                            self.send_response(404)
                            self.send_header("Content-type", "text/plain")
                            self.end_headers()
                            self.wfile.write(b"404 invalid track position")
                    else:
                        self.send_response(404)
                        self.send_header("Content-type", "text/plain")
                        self.end_headers()
                        self.wfile.write(b"404 playlist not found")
                else:
                    self.send_response(404)
                    self.send_header("Content-type", "text/plain")
                    self.end_headers()
                    self.wfile.write(b"404 invalid track")

            elif path.startswith("/api1/setvolume/"):
                key = path[16:]
                if key.isdigit():
                    volume = int(key)
                    volume = max(volume, 0)
                    volume = min(volume, 100)
                    pctl.player_volume = volume
                    pctl.set_volume()

                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                self.wfile.write(b"OK")

            elif path.startswith("/api1/setvolumerel/"):
                key = path[19:]
                if key.lstrip("-").isdigit() and key.count("-") < 2:
                    volume = pctl.player_volume + int(key)
                    volume = max(volume, 0)
                    volume = min(volume, 100)
                    pctl.player_volume = volume
                    pctl.set_volume()

                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                self.wfile.write(b"OK")

            elif path.startswith("/api1/seek1k/"):
                key = path[13:]
                if key.isdigit():
                    pctl.seek_decimal(int(key) / 1000)

                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                self.wfile.write(b"OK")

            elif path.startswith("/api1/seek/"):
                key = path[11:]
                if key.isdigit():
                    pctl.seek_time(int(key) / 1000)

                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                self.wfile.write(b"OK")

            elif path.startswith("/api1/tracklist/"):
                # Return all tracks in a playlist /tracklist/plid
                key = path[16:]
                l = []
                if key.isdigit():
                    pl = tauon.id_to_pl(int(key))
                    if pl is not None and pl < len(pctl.multi_playlist):
                        playlist = pctl.multi_playlist[pl][2]
                        parent = ""
                        album_id = 0
                        for i, id in enumerate(playlist):
                            tr = pctl.g(id)
                            if i == 0:
                                parent = tr.parent_folder_path
                            elif parent != tr.parent_folder_path:
                                parent = tr.parent_folder_path
                                album_id = i

                            l.append(self.get_track(i, pl, album_id=album_id))

                data = {"tracks": l}
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                data = json.dumps(data).encode()
                self.wfile.write(data)

            elif path.startswith("/api1/albums/"):
                # Returns lists of tracks that are start of albums /albums/plid
                key = path[13:]
                l = []
                if key.isdigit():
                    pl = tauon.id_to_pl(int(key))
                    if pl is not None:
                        dex = tauon.reload_albums(True, pl)
                        # print(dex)
                        for a in dex:
                            l.append(self.get_track(a, pl, album_id=a))

                data = {"albums": l}
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                data = json.dumps(data).encode()
                self.wfile.write(data)

            elif path == "/api1/status":
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                data = {
                    "status": "stopped",
                    "inc": pctl.db_inc,
                    "shuffle": pctl.random_mode == True,
                    "repeat": pctl.repeat_mode == True,
                    "progress": 0,
                    "volume": pctl.player_volume,
                    "playlist": str(tauon.get_playing_playlist_id()),
                    "playlist_length": len(pctl.multi_playlist[pctl.active_playlist_playing][2])
                }
                if pctl.playing_state == 1:
                    data["status"] = "playing"
                if pctl.playing_state == 2:
                    data["status"] = "paused"
                track = pctl.playing_object()
                if track:
                    data["id"] = track.index
                    data["title"] = track.title
                    data["artist"] = track.artist
                    data["album"] = track.album
                    data["progress"] = int(round(pctl.playing_time * 1000))
                    data["track"] = self.get_track(0, 0, track)

                p = pctl.playlist_playing_position
                data["position"] = p
                data["album_id"] = 0
                playlist = pctl.playing_playlist()

                if p < len(playlist):
                    while True:
                        if p < 0 or pctl.g(playlist[p]).parent_folder_path != track.parent_folder_path:
                            p += 1
                            break
                        p -= 1
                    data["album_id"] = p

                data = json.dumps(data).encode()
                self.wfile.write(data)

            else:
                self.send_response(404)
                self.end_headers()
                self.wfile.write(b"404 Not found")
            tauon.wake()

    class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
        pass

    try:
        httpd = ThreadedHTTPServer(("0.0.0.0", 7814), Server)
        httpd.serve_forever()
        httpd.server_close()
    except OSError:
        print("Not starting web api server, already running?")


def controller(tauon):
    import base64
    class Server(BaseHTTPRequestHandler):
        def do_GET(self):

            path = self.path
            if path == "/raise/":
                tauon.raise_window()
            if path == "/playpause/":
                if tauon.pctl.playing_state == 0:
                    tauon.pctl.play()
                else:
                    tauon.pctl.pause()
            if path == "/play/":
                tauon.pctl.play()
            if path == "/pause/":
                tauon.pctl.pause_only()
            if path == "/stop/":
                tauon.pctl.stop()
            if path == "/next/":
                tauon.pctl.advance()
            if path == "/previous/":
                tauon.pctl.back()
            if path == "/shuffle/":
                tauon.toggle_random()
            if path == "/repeat/":
                tauon.toggle_repeat()
            if path.startswith("/open/"):
                rest = path[6:]
                path = base64.urlsafe_b64decode(rest.encode()).decode()
                tauon.open_uri(path)

            self.send_response(200)
            self.end_headers()
            tauon.wake()
    print("Start controller server")
    try:
        httpd = HTTPServer(("127.0.0.1", 7813), Server)
        httpd.serve_forever()
        httpd.server_close()
    except OSError:
        print("Not starting controller server, already running?")


def authserve(tauon):

    class Server(BaseHTTPRequestHandler):

        def do_GET(self):
            code = ""
            path = self.path
            if path.startswith("/spotredir"):
                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                code = path.split("code=")
                if len(code) > 1:
                    code = code[1]
                    self.wfile.write(b"You can close this now and return to Tauon Music Box")
                tauon.wake()

            else:
                self.send_response(400)
                self.end_headers()

            if code:
                tauon.spot_ctl.paste_code(code)

    httpd = HTTPServer(("127.0.0.1", 7811), Server)
    httpd.serve_forever()
    httpd.server_close()

import struct
class VorbisMonitor():

    def __init__(self):

        self.tauon = None
        self.reset()
        self.enable = True
        self.synced = False
        self.buffer = io.BytesIO()
        self.tries = 0

    def reset(self, tries=0):
        self.enable = True
        self.synced = False
        self.buffer = io.BytesIO()
        self.tries = tries

    def input(self, data):

        if not self.enable:
            return

        b = self.buffer
        b.seek(0, io.SEEK_END)
        b.write(data)

        # Check theres enough data to decode header
        b.seek(0, io.SEEK_END)
        l = b.tell()
        if l < 128:
            print("Not enough data to parse vorbis")
            return

        # Get page length
        b.seek(0, io.SEEK_SET)
        ogg = b.read(4)

        if not ogg == b"OggS":
            f = data.find(b"Oggs")
            self.reset(self.tries)
            if f > -1:
                print("Ogg stream synced")
                data = data[f:]
                b = self.buffer
                b.write(data)
            else:
                self.tries += 1
                if self.tries > 100:
                    print("Giving up looking for OGG pages")
                    self.enable = False
                return

            # self.enable = False
            # return

        b.seek(0, io.SEEK_SET)
        header = struct.unpack('<4sBBqIIiB', b.read(27))
        segs = struct.unpack('B' * header[7], b.read(header[7]))

        length = 0
        for s in segs:
            length += s

        length += 27 + header[7]

        if l > length:
            h = b.read(7)
            # print(h)
            if h == b"\x03vorbis" or h == b"OpusTag":
                if h == b"OpusTag":
                    b.seek(1, io.SEEK_CUR)

                vendor_length = int.from_bytes(b.read(4), byteorder='little')
                vendor = int.from_bytes(b.read(vendor_length), byteorder='little')
                comment_list_length = int.from_bytes(b.read(4), byteorder='little')

                found_tags = {}
                for i in range(comment_list_length):
                    comment_length = int.from_bytes(b.read(4), byteorder='little')
                    comment = b.read(comment_length)

                    key, value = comment.decode().split("=", 1)

                    if key == "title":
                        found_tags["title"] = value
                    if key == "artist":
                        found_tags["artist"] = value
                    if key == "year":
                        found_tags["year"] = value
                    if key == "album":
                        found_tags["album"] = value

                line = ""
                if "title" in found_tags:
                    line += found_tags["title"]
                    if "artist" in found_tags:
                        line = found_tags["artist"] + " - " + line

                if self.tauon.radiobox.parse_vorbis_okay():
                    self.tauon.pctl.found_tags = found_tags
                    self.tauon.pctl.tag_meta = line

                print("Found vorbis comment")
                print(line)

            # Consume page from buffer
            b.seek(length, io.SEEK_SET)
            new = io.BytesIO()
            new.write(b.read())
            self.buffer = new

vb = VorbisMonitor()

def stream_proxy(tauon):

    class Server(BaseHTTPRequestHandler):

        def do_GET(self):

            self.send_response(200)
            self.send_header("Content-type", "audio/ogg")
            self.end_headers()

            position = 0
            vb.reset()
            vb.tauon = tauon

            while True:
                if not tauon.stream_proxy.download_running:
                    return

                while position < tauon.stream_proxy.c:
                    if position not in tauon.stream_proxy.chunks:
                        print("The buffer was deleted too soon!")
                        return

                    self.wfile.write(tauon.stream_proxy.chunks[position])

                    if tauon.prefs.backend == 4:
                        vb.input(tauon.stream_proxy.chunks[position])

                    position += 1

                else:
                    time.sleep(0.01)

    httpd = HTTPServer(("127.0.0.1", 7812), Server)
    httpd.serve_forever()
    httpd.server_close()




# Tauon Music Box - URL stream download and encoding module

# Copyright © 2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.
import mimetypes
import urllib.request
import threading
import time
import subprocess
import os
import fcntl
import datetime
import io
import shutil
import mutagen
import copy
from PIL import Image
from mutagen.flac import Picture
import base64
from t_modules.t_extra import filename_safe
from t_modules.t_webserve import vb

class StreamEnc:

    def __init__(self, tauon):
        self.tauon = tauon
        self.download_running = False
        self.encode_running = False
        self.pump_running = False
        self.feed_running = False

        self.download_process = False
        self.abort = False

        self.s_name = ""
        self.s_bitrate = ""
        self.s_genre = ""
        self.s_description = ""
        self.s_mime = ""
        self.s_format = ""

        self.chunks = {}
        self.c = 0

    def stop(self):

        try:
            self.tauon.radiobox.websocket.close()
            print("Websocket closed")
        except:
            print("No socket to close?")

        self.abort = True
        self.tauon.radiobox.loaded_url = None

    def start_download(self, url):

        self.abort = True
        while self.download_running:
            time.sleep(0.01)
        while self.encode_running:
            time.sleep(0.01)
        while self.pump_running:
            time.sleep(0.01)

        self.__init__(self.tauon)

        def NiceToICY(self):
            class InterceptedHTTPResponse:
                pass

            line = self.fp.readline().replace(b"ICY 200 OK\r\n", b"HTTP/1.0 200 OK\r\n")
            InterceptedSelf = InterceptedHTTPResponse()
            InterceptedSelf.fp = io.BufferedReader(io.BytesIO(line))
            InterceptedSelf.debuglevel = self.debuglevel
            InterceptedSelf._close_conn = self._close_conn
            return ORIGINAL_HTTP_CLIENT_READ_STATUS(InterceptedSelf)

        ORIGINAL_HTTP_CLIENT_READ_STATUS = urllib.request.http.client.HTTPResponse._read_status
        urllib.request.http.client.HTTPResponse._read_status = NiceToICY

        try:
            r = urllib.request.Request(url)
            #r.add_header('GET', '1')
            r.add_header('Icy-MetaData', '1')
            r.add_header('User-Agent', self.tauon.t_agent)
            print("Open URL.....")
            r = urllib.request.urlopen(r, timeout=7)
            print("URL opened.")

        except Exception as e:
            print("Connection failed")
            #raise
            self.tauon.gui.show_message("Failed to establish a connection", str(e), mode="error")

            return False

        self.download_process = threading.Thread(target=self.run_download, args=([r]))
        self.download_process.daemon = True
        self.download_process.start()
        self.download_running = True

        self.download_process = threading.Thread(target=self.pump)
        self.download_process.daemon = True
        self.download_process.start()
        return True

    def pump(self):
        aud = self.tauon.aud
        if self.tauon.prefs.backend != 4 or not aud:
            print("Radio error: Phazor not loaded")
            return
        self.pump_running = True

        rate = str(48000)
        cmd = ['ffmpeg', "-loglevel", "quiet", "-i", "pipe:0", "-acodec", "pcm_f32le", "-f", "f32le", "-ac", "2", "-ar",
               rate, "-"]

        decoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        fcntl.fcntl(decoder.stdout.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)

        raw_audio = None
        max_read = int(10000)
        vb.reset()
        vb.tauon = self.tauon

        def feed(decoder):
            position = 0
            self.feed_running = True
            try:
                while True:
                    if position < self.tauon.stream_proxy.c:
                        if position not in self.tauon.stream_proxy.chunks:
                            print("The buffer was deleted too soon!")
                            break

                        chunk = self.chunks[position]
                        decoder.stdin.write(chunk)
                        vb.input(self.tauon.stream_proxy.chunks[position])
                        position += 1
                    else:
                        time.sleep(0.01)
                    if not self.pump_running or not self.feed_running:
                        break
            except:
                self.feed_running = False
                raise
            print("Exit feeder")

        feeder = threading.Thread(target=feed, args=[decoder])
        feeder.daemon = True
        feeder.start()

        while True:
            if not self.tauon.stream_proxy.download_running or self.abort:
                break

            if raw_audio == None:
                raw_audio = decoder.stdout.read(max_read)
            if raw_audio:
                r = aud.feed_ready(max_read)
                if r:
                    aud.feed_raw(len(raw_audio), raw_audio)
                    raw_audio = None
                else:
                    time.sleep(0.01)
                    continue

        decoder.terminate()
        time.sleep(0.1)
        try:
            decoder.kill()
        except:
            pass

        self.pump_running = False


    def encode(self):

        self.encode_running = True

        try:

            while self.c < 20:
                if self.abort:
                    self.encode_running = False
                    return
                time.sleep(0.05)

            ext = ".opus"
            rate = "48000"
            codec = self.tauon.prefs.radio_record_codec.upper()
            if codec == "OGG":
                ext = ".ogg"
                rate = "44100"
            if codec == "MP3":
                ext = ".mp3"
                rate = "44100"
            if codec == "FLAC":
                ext = ".flac"
                rate = "44100"

            target_file = os.path.join(self.tauon.cache_directory, "stream" + ext)
            if os.path.isfile(target_file):
                os.remove(target_file)

            cmd = ['ffmpeg', "-loglevel", "quiet", "-i", "pipe:0", "-acodec", "pcm_s16le", "-f", "s16le", "-ac", "2", "-ar", rate, "-"]

            decoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            fcntl.fcntl(decoder.stdout.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)

            position = 0
            old_metadata = self.tauon.radiobox.song_key
            old_tags = self.tauon.pctl.found_tags

            ##cmd = ["opusenc", "--raw", "--raw-rate", "48000", "-", target_file]
            cmd = ["ffmpeg", "-loglevel", "quiet", "-f", "s16le", "-ar", rate, "-ac", "2", "-i", "pipe:0", target_file]
            encoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

            def save_track():
                #self.tauon.recorded_songs.append(song)

                save_file = '{:%Y-%m-%d %H-%M-%S} - '.format(datetime.datetime.now())
                save_file += filename_safe(old_metadata)
                save_file = save_file.strip() + ext
                save_file = os.path.join(self.tauon.prefs.encoder_output, save_file)
                if os.path.exists(save_file):
                    os.remove(save_file)
                if not os.path.exists(self.tauon.prefs.encoder_output):
                    os.makedirs(self.tauon.prefs.encoder_output)
                shutil.move(target_file, save_file)

                # print(self.tauon.pctl.tag_history)
                # print(old_metadata)
                tags = self.tauon.pctl.tag_history.get(old_metadata, None)
                if tags:
                    print("Save metadata to file")
                    #print(tags)
                    muta = mutagen.File(save_file, easy=True)
                    muta["artist"] = tags.get("artist", "")
                    muta["title"] = tags.get("title", "")
                    muta["album"] = tags.get("album", "")
                    # if tags["image"]:
                    #     tags["image"].seek(0)
                    #     im = Image.open(tags["image"])
                    #     width, height = im.size
                    #     tags["image"].seek(0)
                    #
                    #     picture = Picture()
                    #     tags["image"].seek(0)
                    #     picture.data = tags["image"].read()
                    #     picture.type = 3
                    #     picture.desc = ""
                    #     picture.mime = "image/jpeg"
                    #     picture.width = width
                    #     picture.height = height
                    #
                    #     mode_to_bpp = {'1': 1, 'L': 8, 'P': 8, 'RGB': 24, 'RGBA': 32, 'CMYK': 32, 'YCbCr': 24, 'I': 32,
                    #                    'F': 32}
                    #     picture.depth = mode_to_bpp[im.mode]
                    #
                    #     picture_data = picture.write()
                    #     encoded_data = base64.b64encode(picture_data)
                    #     vcomment_value = encoded_data.decode("ascii")
                    #     muta["metadata_block_picture"] = [vcomment_value]

                    muta.save()

                target_pl = None
                for i, pl in enumerate(self.tauon.pctl.multi_playlist):
                    if pl[0] == "Saved Radio Tracks":
                        target_pl = i
                if target_pl is None:
                    self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title="Saved Radio Tracks"))
                    target_pl = len(self.tauon.pctl.multi_playlist) - 1

                load_order = self.tauon.pctl.LoadClass()
                load_order.playlist = self.tauon.pctl.multi_playlist[target_pl][6]
                load_order.target = save_file
                self.tauon.load_orders.append(copy.deepcopy(load_order))
                self.tauon.gui.update += 1

            while True:

                if self.abort:
                    decoder.terminate()
                    encoder.terminate()
                    time.sleep(0.1)
                    try:
                        decoder.kill()
                    except:
                        pass
                    try:
                        encoder.kill()
                    except:
                        pass

                    if os.path.exists(target_file):
                        if os.path.getsize(target_file) > 256000:
                            print("Save file")
                            save_track()
                        else:
                            print("Discard small file")
                            os.remove(target_file)

                    self.encode_running = False
                    self.tauon.pctl.tag_history.clear()
                    return

                if old_metadata != self.tauon.radiobox.song_key:
                    if self.c < 400 and not old_metadata:
                        old_metadata = self.tauon.radiobox.song_key
                    elif not os.path.exists(target_file) or os.path.getsize(target_file) < 100000:
                        old_metadata = self.tauon.radiobox.song_key
                    else:
                        print("Split and save file")
                        encoder.stdin.close()
                        try:
                            encoder.wait(timeout=4)
                        except:
                            pass
                        try:
                            encoder.kill()
                        except:
                            pass
                        if os.path.exists(target_file):
                            if os.path.getsize(target_file) > 256000:
                                save_track()
                            else:
                                print("Discard small file")
                                os.remove(target_file)
                        encoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

                raw_audio = decoder.stdout.read(1000000)
                if raw_audio:
                    encoder.stdin.write(raw_audio)

                if position < self.c:
                    chunk = self.chunks[position]
                    position += 1
                    decoder.stdin.write(chunk)
                else:
                    time.sleep(0.005)

        except Exception as e:
            print("Encoder thread crashed!")
            print(str(e))
            #raise
            self.encode_running = False
            return

    def run_download(self, r):

        h = r.info()

        self.s_name = h.get("icy-name")
        metaint = h.get("icy-metaint")
        self.s_bitrate = h.get("icy-br")
        self.s_genre = h.get("icy-genre")
        self.s_description = h.get("icy-description")
        self.s_mime = h.get("Content-Type")

        print(self.s_mime)
        if self.s_mime == "audio/mpeg":
            self.s_format = "MP3"
        if self.s_mime == "audio/ogg":
            self.s_format = "OGG"
        if self.s_mime == "audio/aac":
            self.s_format = "AAC"
        if self.s_mime == "audio/aacp":
            self.s_format = "AAC+"

        test_done = 0

        icy = False
        m_remain = 0
        m = 0
        if metaint and int(metaint) > 0:
            m = int(metaint)
            m_remain = m
            icy = True

        maybe = b""

        if self.tauon.prefs.auto_rec:
            self.download_process = threading.Thread(target=self.encode)
            self.download_process.daemon = True
            self.download_process.start()

        try:
            while True:

                chunk = r.read(256)

                if self.abort:
                    r.close()
                    print("Abort stream connection")
                    self.download_running = False
                    return

                if chunk:
                    if not icy or m_remain > len(chunk):
                        # We're sure its data Its data, send it on
                        self.chunks[self.c] = chunk

                        # Delete old data
                        d = self.c - 90000
                        if d in self.chunks:
                            del self.chunks[d]

                        test_done += len(chunk)
                        self.c += 1
                        m_remain -= len(chunk)

                        continue
                    else:
                        # It may contain the metadata block, put it aside
                        maybe += chunk

                # Try to extract ICY tag
                if maybe:
                    data1 = maybe[:m_remain]
                    inter = maybe[m_remain:]

                    # Read the metadata length byte
                    if inter:
                        special = inter[0]
                        follow = special * 16

                        if len(inter) < follow + 2:
                            # Not enough data yet
                            continue

                        text = inter[1:follow + 1]
                        data2 = inter[follow + 1:]

                        self.chunks[self.c] = data1 + data2
                        # Delete old data
                        d = self.c - 90000
                        if d in self.chunks:
                            del self.chunks[d]

                        self.c += 1

                        test_done += len(data1)
                        test_done = 0

                        m_remain = m - len(data2)
                        test_done += len(data2)
                        maybe = b""

                        try:
                            meta = text.decode().rstrip("\x00")
                            for tag in meta.split(";"):
                                if '=' in tag:
                                    a, b = tag.split('=', 1)
                                    if a == 'StreamTitle':
                                        #print("Set meta")
                                        self.tauon.pctl.tag_meta = b.rstrip("'").lstrip("'")
                                        break
                        except:
                            r.close()
                            self.download_running = False
                            self.tauon.gui.show_message("Data malformation detected. Stream aborted.", mode='error')
                            raise
        except:
            print("Stream download thread crashed!")
            self.download_running = False
            return



import pychromecast
from t_modules.t_extra import shooter
import time
import socket

def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP


class Chrome:
    def __init__(self, tauon):
        self.tauon = tauon
        self.services = []
        self.active = False
        self.cast = None
        self.target_playlist = None
        self.target_id = None

    def two(self):
        print("Scanning for chromecasts...")

        if True: #not self.services:
            try:
                self.tauon.gui.show_message(self.tauon.strings.scan_chrome)
                services, browser = pychromecast.discovery.discover_chromecasts()
                pychromecast.discovery.stop_discovery(browser)
                menu = self.tauon.chrome_menu
                menu.items.clear()
                for item in services:
                    self.services.append([str(item.uuid), str(item.friendly_name)])
                    menu.add(self.tauon.strings.cast_to % str(item.friendly_name), self.three, pass_ref=True, set_ref=[str(item.uuid), str(item.friendly_name)])

                if self.services:
                    self.tauon.gui.message_box = False
                    menu.activate(position=(self.tauon.gui.window_size[0] // 2, self.tauon.gui.window_size[1] // 2))
                    self.tauon.gui.update += 1
                else:
                    self.tauon.gui.show_message(self.tauon.strings.no_chromecasts)
                    self.tauon.gui.update += 1

                    return
                print(services)
                print(self.services)

            except:
                raise
                print("Failed to get chromecasts")

    def one(self, playlist_id, track_id):

        self.target_playlist = playlist_id
        self.target_id = track_id
        self.cast = None
        if self.cast is None:
            self.two()
            return

    def three(self, item):
        shooter(self.four, [item])

    def four(self, item):

        self.tauon.broadcast_select_track(self.target_id)
        ip = get_ip()
        time.sleep(2)
        print(item)
        ccs, browser = pychromecast.get_listed_chromecasts(friendly_names=[item[1]], discovery_timeout=3.0)
        print(ccs)
        self.cast = ccs[0]
        self.cast.wait()

        mc = self.cast.media_controller
        mc.play_media(f"http://{ip}:7590/stream.ogg", "audio/ogg")
        self.active = True
        pychromecast.discovery.stop_discovery(browser)


    def stop(self):
        if self.active:
            if self.cast:
                mc = self.cast.media_controller
                mc.stop()
            self.active = False




# Tauon Music Box - System cursor loader

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.


import gi
gi.require_version('Gdk', '3.0')
from gi.repository import Gdk
from sdl2 import *
import ctypes


def cursor_get_gdk(type):

    if type == 4:
        cursor = Gdk.Cursor(Gdk.CursorType.BOTTOM_RIGHT_CORNER)
    elif type == 8:
        cursor = Gdk.Cursor(Gdk.CursorType.RIGHT_SIDE)
    elif type == 9:
        cursor = Gdk.Cursor(Gdk.CursorType.TOP_SIDE)
    elif type == 10:
        cursor = Gdk.Cursor(Gdk.CursorType.LEFT_SIDE)
    elif type == 11:
        cursor = Gdk.Cursor(Gdk.CursorType.BOTTOM_SIDE)

    else:
        return None

    return Gdk.Cursor.get_surface(cursor)


def cairo_cursor_to_sdl(cairo_surface, x_hot, y_hot, fallback=None):

    if not cairo_surface:
        return fallback

    w = cairo_surface.get_width()
    h = cairo_surface.get_height()

    mem = cairo_surface.get_data()
    buff = ctypes.c_buffer(mem.tobytes())

    sdl_surface = SDL_CreateRGBSurfaceWithFormatFrom(ctypes.pointer(buff), w, h, 32, w * 4, SDL_PIXELFORMAT_ARGB8888)
    return SDL_CreateColorCursor(sdl_surface, round(x_hot), round(y_hot))




# Tauon Music Box - Tag Module

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.


# The purpose of this module is to read metadata from FLAC, OGG, OPUS, APE and WV files

import struct
import wave
import io
import os
import math
from t_modules.t_extra import process_odat

def parse_mbids_from_vorbis(object, key, value):

    if key == "musicbrainz_artistid":
        if "musicbrainz_artistids" not in object.misc:
            object.misc['musicbrainz_artistids'] = []
        object.misc['musicbrainz_artistids'].append(value)
        return True

    if key == "musicbrainz_trackid":
        object.misc['musicbrainz_recordingid'] = value
        return True

    if key == "musicbrainz_releasetrackid":
        object.misc['musicbrainz_trackid'] = value
        return True

    if key == "musicbrainz_albumid":
        object.misc['musicbrainz_albumid'] = value
        return True

    if key == "musicbrainz_releasegroupid":
        object.misc['musicbrainz_releasegroupid'] = value
        return True

    return False


def parse_picture_block(f):
    a = f.read(4)
    a = int.from_bytes(a, byteorder='big')
    # print("Picture type: " + str(a))

    a = f.read(4)
    b = int.from_bytes(a, byteorder='big')
    # print("MIME len: " + str(b))

    a = f.read(b)
    # print(a)
    # print("MIME: " + a.decode('ascii'))

    a = f.read(4)
    a = int.from_bytes(a, byteorder='big')
    # print("Description len: " + str(a))

    a = f.read(a)
    # print("Description: " + a.decode('utf-8'))

    a = f.read(4)
    # a = int.from_bytes(a, byteorder='big')
    # print("Width: " + str(a))

    a = f.read(4)
    # a = int.from_bytes(a, byteorder='big')
    # print("Height: " + str(a))

    a = f.read(4)
    # a = int.from_bytes(a, byteorder='big')
    # print("BPP: " + str(a))

    a = f.read(4)
    # a = int.from_bytes(a, byteorder='big')
    # print("Index colour: " + str(a))

    a = f.read(4)
    a = int.from_bytes(a, byteorder='big')
    # print("Bin len: " + str(a))

    return f.read(a)

class Flac:

    def __init__(self, file):

        self.filepath = file
        self.has_picture = False
        self.picture = ""

        self.album_artist = ""
        self.artist = ""
        self.genre = ""
        self.date = ""
        self.comment = ""
        self.album = ""
        self.track_number = ""
        self.track_total = ""
        self.title = ""
        self.encoder = ""
        self.disc_number = ""
        self.picture = ""
        self.disc_total = ""
        self.lyrics = ""
        self.cue_sheet = ""
        self.composer = ""
        self.misc = {}

        self.sample_rate = 48000
        self.bit_rate = 0
        self.length = 0
        self.bit_depth = 0

        self.track_gain = None
        self.album_gain = None


    def read_vorbis(self, f):

        block_position = 0

        buffer = f.read(4)
        block_position += 4
        jump = int.from_bytes(buffer, byteorder='little')

        f.read(jump)
        block_position += jump

        buffer = f.read(4)
        block_position += 4
        fields = int.from_bytes(buffer, byteorder='little')
        # print(fields)
        artists = []
        genres = []
        odat = ""

        for i in range(fields):
            buffer = f.read(4)
            block_position += 4

            jump = int.from_bytes(buffer, byteorder='little')
            buffer = f.read(jump)
            block_position += jump

            # print(buffer.decode('utf-8'))

            position = 0

            while position < 40:
                # print(sss[position:position+1])
                position += 1


                if buffer[position:position + 1] == b'=':

                    a = buffer[0:position].decode("utf-8").lower()
                    b = buffer[position + 1:]

                    if parse_mbids_from_vorbis(self, a, b.decode()):
                        pass
                        # print("Found MBID data:")
                        # print(a)
                        # print(b)
                    elif a == "genre":
                        #self.genre = b.decode("utf-8")
                        genres.append(b.decode())
                    elif a == 'cuesheet':
                        self.cue_sheet = b.decode()
                    elif a == "date":
                        self.date = b.decode("utf-8")
                    elif a == "originaldate":
                        odat = b.decode("utf-8")
                    elif a == "comment":
                        self.comment = b.decode("utf-8")
                    elif a == "album":
                        self.album = b.decode("utf-8")
                    elif a == "title":
                        self.title = b.decode("utf-8")
                    elif a == "tracknumber":
                        self.track_number = b.decode("utf-8")
                    elif a == "tracktotal" or a == "totaltracks":
                        self.track_total = b.decode("utf-8")
                    elif a == "encoder":
                        self.encoder = b.decode("utf-8")
                    elif a == "albumartist" or a == "album artist":
                        self.album_artist = b.decode("utf-8")
                    elif a == "artist":
                        #self.artist = b.decode("utf-8")
                        artists.append(b.decode())
                    elif a == 'disctotal' or a == 'totaldiscs':
                        self.disc_total = b.decode("utf-8")
                    elif a == "discnumber":
                        self.disc_number = b.decode("utf-8")
                    elif a == "metadata_block_picture":
                        print("Tag Scanner: Found picture inside vorbis comment inside a FLAC file. Ignoring")
                        print("      In file: " + self.filepath)
                    elif a == 'lyrics' or a == 'unsyncedlyrics':
                        self.lyrics = b.decode("utf-8")
                    elif "replaygain_track_gain" == a:
                        self.misc["replaygain_track_gain"] = float(b.decode("utf-8").strip(" dB"))
                    elif "replaygain_track_peak" == a:
                        self.misc["replaygain_track_peak"] = float(b.decode("utf-8"))
                    elif "replaygain_album_gain" == a:
                        self.misc["replaygain_album_gain"] = float(b.decode("utf-8").strip(" dB"))
                    elif "replaygain_album_peak" == a:
                        self.misc["replaygain_album_peak"] = float(b.decode("utf-8"))
                    elif 'composer' == a:
                        self.composer = b.decode("utf-8")
                    elif "fmps_rating" == a:
                        self.misc['FMPS_Rating'] = float(b.decode("utf-8"))
                    # else:
                    #     print("Tag Scanner: Found unhandled FLAC Vorbis comment field: " + a)
                    #     print(b)
                    #     print("\n-------------------------------------------\n")

        f.seek(block_position * -1, 1)

        if artists:
            self.artist = "; ".join(artists)
            if len(artists) > 1:
                self.misc['artists'] = artists
        if genres:
            self.genre = "; ".join(genres)
            if len(genres) > 1:
                self.misc['genres'] = genres
        process_odat(self, odat)

    def read_seek_table(self, f):

        f.read(10)
        buffer = f.read(8)
        a = (int.from_bytes(buffer, byteorder='big'))
        k = bin(a)[2:].zfill(64)

        self.sample_rate = int(k[0:20], 2)

        bps = int(k[23:28], 2)
        self.bit_depth = bps + 1

        samples = int(k[28:64], 2)

        self.length = samples / self.sample_rate
        f.seek(-18, 1)

    def read(self, get_picture=False):

        # Very helpful: https://xiph.org/flac/format.html
        size = os.path.getsize(self.filepath) / 8
        if size < 100:
            return

        f = open(self.filepath, "rb")
        s = f.read(4)

        # Find start of FLAC stream
        if s != b'fLaC':
            while f.tell() < size + 100:
                f.seek(-3, 1)
                s = f.read(4)
                if s == b'fLaC':
                    break
            else:
                return

        i = 0
        while i < 20:
            i += 1

            z = self.read_block(f)

            if z[1] == 0:
                self.read_seek_table(f)
            if z[1] == 4:
                self.read_vorbis(f)

            if z[1] == 5:

                print("Tag Scan: Flac file has native embedded CUE. Not supported")
                print("      In file: " + self.filepath)
                # mark = f.tell()
                #
                # print("Found flac cue")
                #
                # a = f.read(16*8)
                # a = int.from_bytes(a, byteorder='big')
                # print(("Catalog Number: " + str(a)))
                #
                #
                # f.seek(mark)
                # f.seek(z[2], 1)

            if z[1] == 6 and get_picture:
                self.picture = parse_picture_block(f)
                self.has_picture = True

            else:
                f.read(z[2])

            if z[0] == 1:
                break

        f.close()

    def read_block(self, f):

        q = f.read(1)
        a = (int.from_bytes(q, byteorder='big'))
        k = bin(a)[2:].zfill(8)
        flag = int(k[:1], 2)
        block_type = int(k[1:], 2)
        s = f.read(3)
        a = (int.from_bytes(s, byteorder='big'))
        length = a

        return flag, block_type, length

    def get(self):
        pass


# file = 'b.flac'
#
# item = Flac(file)
# item.read()


class Opus:

    def __init__(self, file):

        self.filepath = file
        self.has_picture = False

        self.album_artist = ""
        self.artist = ""
        self.genre = ""
        self.date = ""
        self.comment = ""
        self.album = ""
        self.track_number = ""
        self.track_total = ""
        self.title = ""
        self.encoder = ""
        self.disc_number = ""
        self.disc_total = ""
        self.picture = ""
        self.lyrics = ""
        self.composer = ""
        self.track_gain = None
        self.album_gain = None
        self.misc = {}

        self.sample_rate = 48000  # OPUS files are always 48000
        self.bit_rate = 0
        self.length = 0

    def get_more(self, f, v):

        header = struct.unpack('<4sBBqIIiB', f.read(27))

        segs = struct.unpack('B' * header[7], f.read(header[7]))
        l = sum(segs)

        o = v.tell()
        v.seek(0, 2)
        v.write(f.read(l))
        v.seek(o)

        return l

    def read(self):

        f = open(self.filepath, "rb")

        header = struct.unpack('<4sBBqIIiB', f.read(27))

        # print(header)

        segs = struct.unpack('B'*header[7], f.read(header[7]))

        # l = sum(segs)
        # print(f.read(l + 4))
        # f.seek(l * -1)

        s = f.read(7)

        if s == b'OpusHea':
            f.seek(-7, 1)
        elif s == b'\x01vorbis':

            s = f.read(4)
            a = struct.unpack("<B4i", f.read(17)) # 44100
            self.sample_rate = a[1]
            self.bit_rate = int(a[3] / 1000)
            f.seek(-28, 1)

        else:
            return

        for p in segs:
            f.read(p)

        v = io.BytesIO()
        v.seek(0)

        l = self.get_more(f, v)

        s = v.read(7)
        l -= 7

        if s == b"OpusTag":
            v.read(1)
            l -= 1
        elif s == b"\x03vorbis":
            pass
        else:
            return

        s = v.read(4)
        l -= 4
        a = int.from_bytes(s, byteorder='little')  # Vendor string length
        s = v.read(a)  # Vendor string
        l -= a

        s = v.read(4)
        l -= 4

        number = int.from_bytes(s, byteorder='little')  # Number of comments

        artists = []
        genres = []
        odat = ""

        for i in range(number):
            s = v.read(4)
            l -= 4

            length = int.from_bytes(s, byteorder='little')

            while l < length + 4:
                l += self.get_more(f, v)

            s = v.read(length)
            l -= length

            position = 0
            while position < 40:

                position += 1

                if s[position:position+1] == b'=':

                    a = s[0:position].decode("utf-8").lower()
                    b = s[position + 1:]

                    # print(a)  # Key
                    # print(b)  # Value

                    if parse_mbids_from_vorbis(self, a, b.decode()):
                        pass
                        # print("Found MBID data:")
                        # print(a)
                        # print(b)

                    elif a == "genre":
                        #self.genre = b.decode("utf-8")
                        genres.append(b.decode())
                    elif a == "date":
                        self.date = b.decode("utf-8")
                    elif a == "originaldate":
                        odat = b.decode("utf-8")
                    elif a == "comment":
                        self.comment = b.decode("utf-8")
                    elif a == "album":
                        self.album = b.decode("utf-8")
                    elif a == "title":
                        self.title = b.decode("utf-8")
                    elif a == "tracknumber":
                        self.track_number = b.decode("utf-8")
                    elif a == "tracktotal" or a == "totaltracks":
                        self.track_total = b.decode("utf-8")
                    elif a == "encoder":
                        self.encoder = b.decode("utf-8")
                    elif a == "albumartist" or a == "album artist":
                        self.album_artist = b.decode("utf-8")
                    elif a == "artist":
                        #self.artist = b.decode("utf-8")
                        artists.append(b.decode())
                    elif a == "metadata_block_picture":

                        print("Tag Scanner: Found picture in OGG/OPUS file.")
                        print("      In file: " + self.filepath)
                        self.has_picture = True
                        self.picture = b
                        # print(b)

                    elif 'replaygain_track_gain' == a:
                        self.misc["replaygain_track_gain"] = float(b.decode("utf-8").strip(" dB"))
                    elif 'replaygain_track_peak' == a:
                        self.misc["replaygain_track_peak"] = float(b.decode("utf-8"))
                    elif 'replaygain_album_gain' == a:
                        self.misc["replaygain_album_gain"] = float(b.decode("utf-8").strip(" dB"))
                    elif 'replaygain_album_peak' == a:
                        self.misc["replaygain_album_peak"] = float(b.decode("utf-8"))
                    elif a == "discnumber":
                        self.disc_number = b.decode("utf-8")
                    elif a == 'disctotal' or a == 'totaldiscs':
                        self.disc_total = b.decode("utf-8")
                    elif a == 'lyrics' or a == 'unsyncedlyrics':
                        self.lyrics = b.decode("utf-8")
                    elif a == "composer":
                        self.composer = b.decode("utf-8")
                    elif "fmps_rating" == a:
                        self.misc['FMPS_Rating'] = float(b.decode("utf-8"))
                    # else:
                    #     print("Tag Scanner: Found unhandled Vorbis comment field: " + a)
                    #     print(b.decode("utf-8"))
                    #     print("      In file: " + self.filepath)

                    break

        v.close()

        if artists:
            self.artist = "; ".join(artists)
            if len(artists) > 1:
                self.misc['artists'] = artists
        if genres:
            self.genre = "; ".join(genres)
            if len(genres) > 1:
                self.misc['genres'] = genres
        process_odat(self, odat)

        # Find the last Ogg page from end of file to get track length
        f.seek(-1, 2)

        # Crudely seek back for it
        g = 100000
        while g > 0:
            if f.tell() < 10:
                break
            g -= 1
            f.seek(-5, 1)
            s = f.read(4)

            if s == b"OggS":
                f.seek(-4, 1)
                header = struct.unpack('<4sBBqIIiB', f.read(27))
                self.length = header[3] / self.sample_rate
                break

        f.close()

# file = 'a.ogg'
#
# item = Opus(file)
# item.read()


class Ape:

    def __init__(self, file):

        self.filepath = file
        self.has_picture = False
        self.picture = ""

        self.found_tag = False
        self.album_artist = ""
        self.artist = ""
        self.genre = ""
        self.date = ""
        self.comment = ""
        self.album = ""
        self.track_number = ""
        self.track_total = ""
        self.title = ""
        self.encoder = ""
        self.disc_number = ""
        self.disc_total = ""
        self.picture = ""
        self.lyrics = ""
        self.label = ""
        self.track_gain = None
        self.album_gain = None
        self.misc = {}
        self.composer = ""

        self.sample_rate = 48000
        self.bit_rate = 0
        self.bit_depth = 0
        self.length = 0

    def read(self):

        a = open(self.filepath, 'rb')

        # Check size of file
        a.seek(0, 2)
        file_size = a.tell()

        # Helpful: http://wiki.hydrogenaud.io/index.php?title=APEv2_specification

        # Get last 32 bytes where ape tag footer might be
        a.seek(-32, 1)
        b = a.read(32)
        footer = struct.unpack('<8c6i', b)

        # For use later
        found = 1

        # If its not an ape footer, seek through the file for a bit to see if we find it
        if b"".join(footer[0:8]) != b'APETAGEX':
            found = False
            a.seek(0, 2)
            g = 1000
            while g > 0:
                if a.tell() < 1000:
                    break
                g -= 1
                a.seek(-9, 1)
                s = a.read(8)

                # Found it
                if s == b"APETAGEX":
                    found = 2
                    a.seek(-8, 1)
                    b = a.read(32)
                    footer = struct.unpack('<8c6i', b)

        if found == 0:
            print("Tag Scanner: Cant find APE tag")
        else:

            self.found_tag = True
            tag_len = footer[9]  # The size of the tag data (excludes header)
            num_items = footer[10]  # Number of fields in tag

            # print("Tag len: " + str(tag_len))
            # print("Items: " + str(num_items))

            # Seek to start of tag (after any header)
            if found == 1:
                a.seek(tag_len * -1, 2)
            else:
                a.seek(8 + (tag_len * -1), 1)

            # For every field in tag
            for q in range(num_items):

                # (field value length, 0=text 2=binary)
                ta = struct.unpack("<ii", a.read(8))

                # Collect every character until we reach null terminator
                name = b""
                for i in range(100):
                    ch = a.read(1)
                    if ch == b"\x00":
                        break
                    name += ch

                key = name.decode('utf-8')
                # print("Key: " + key)

                value = a.read(ta[0])
                # print(value)

                if ta[1] == 0:
                    value = value.decode('utf-8')
                elif ta[1] == 2:
                    # Avoid decode of binary data
                    pass

                # Fill in the class attributes
                if key.lower() == "title":
                    self.title = value
                elif key.lower() == "artist":
                    self.artist = value
                elif key.lower() == "genre":
                    self.genre = value
                elif key.lower() == "disc":

                    # Ape track fields appear to use fraction format, rather than separate fields for number and total
                    # So we need to handle that here for consistency
                    if "/" in value:
                        self.disc_number, self.disc_total = value.split('/')
                    else:
                        self.disc_number = value

                elif key.lower() == "comment":
                    self.comment = value
                elif key.lower() == "track":

                    # Same deal as with disc number
                    if "/" in value:
                        self.track_number, self.track_total = value.split('/')
                    else:
                        self.track_number = value

                elif key.lower() == "year":
                    self.date = value
                elif key.lower() == "album":
                    self.album = value
                elif key.lower() == "artist":
                    self.artist = value
                elif key.lower() == "composer":
                    self.composer = value
                elif key.lower() == "album artist":
                    self.album_artist = value
                elif key.lower() == "label":
                    self.label = value
                elif key.lower() == "lyrics":
                    self.lyrics = value
                elif "replaygain_track_gain" == key.lower():
                    self.misc["replaygain_track_gain"] = float(value.strip(" dB"))
                elif "replaygain_track_peak" == key.lower():
                    self.misc["replaygain_track_peak"] = float(value)
                elif "replaygain_album_gain" == key.lower():
                    self.misc["replaygain_album_gain"] = float(value.strip(" dB"))
                elif "replaygain_album_peak" == key.lower():
                    self.misc["replaygain_album_peak"] = float(value)
                elif parse_mbids_from_vorbis(self, key.lower(), value):
                    pass

                elif key.lower() == "cover art (front)":

                    # Data appears to have a filename at the start of it, we need to remove to recover a valid picture
                    # Im not sure what the actual specification is here

                    off = 0
                    while off < 64:
                        if value[off:off+1] == b'\x00':

                            off += 1
                            break
                        off += 1
                    else:
                        print("Tag Scanner: Error reading APE album art")
                        continue

                    self.picture = value[off:]
                    self.has_picture = True
                    # print(value)

        # Back to start of file to see if we can find sample rate and duration information
        a.seek(0)

        start = a.read(128)
        if start[0:3] == b'MAC':  # Ape files start with MAC

            version = struct.unpack("<h", start[4:6])[0]

            if version >= 3980:

                audio_info = struct.unpack("<IIIHHI", start[56:76])

                self.bit_depth = audio_info[3]
                self.sample_rate = audio_info[5]

                frames = audio_info[2] - 1
                blocks = audio_info[0]

                self.length = (frames * blocks) / self.sample_rate
            else:

                print("WARNING: Old APE codec version; not supported")

        elif ".tta" in self.filepath:

            a.seek(0)
            header = struct.unpack("<4c3H3L", a.read(22))

            if b"".join(header[0:3]) != b'TTA1':

                self.sample_rate = header[7]
                bps = header[6]
                self.bit_depth = bps
                # channels = header[5]
                self.length = header[8] / self.sample_rate
            elif b"".join(header[0:3]) != b'TTA2':
                print("WARNING: TTA2 type TTA file not supported")
                # To do
            else:
                print("WARNING: Does not appear to be a valid TTA file")

        elif '.wv' in self.filepath:
            #  We can handle WavPack files here too
            #  This code likely wont cover all cases as is, I only tested it on a few files

            a.seek(0)

            #  I found that some WavPack files have padding at the beginning
            #  So here I crudely search for the actual start
            off = 0
            while off < file_size - 100:
                if a.read(4) == b'wvpk':
                    a.seek(-4, 1)
                    b = a.read(32)
                    header = struct.unpack("<4cIH2B5I", b)

                    sample_rates = [6000, 8000, 9600, 11025, 12000, 16000, 22050, 24000, 32000, 44100, 48000, 64000,
                                    88200, 96000, 192000]   # Adapted from example in WavPack/cli/wvparser.c
                    n = ((header[11] & (15 << 23)) >> 23)   # Does my head in this
                    self.sample_rate = sample_rates[n]
                    self.length = header[8] / self.sample_rate
                    break
            else:
                print("Tag Scanner: Cannot verify WavPack file")

        else:
            print("Tag Scanner: Does not appear to be an APE file")

        a.close()


class Wav:

    def __init__(self, file):

        self.filepath = file
        self.sample_rate = 48000
        self.length = 0

        self.title = ""
        self.artist = ""
        self.album = ""
        self.genre = ""
        self.track_number = ""

    def read(self):

        with open(self.filepath, "rb") as f:
            f.read(12)

            while True:
                type = f.read(4)
                if not type:
                    break
                remain = int.from_bytes(f.read(4), 'little')

                if type != b"LIST":
                    f.seek(remain, io.SEEK_CUR)
                else:

                    INFO = f.read(4)
                    if INFO == b"INFO":
                        remain -= 4
                        while remain > 0:
                            id = f.read(4).decode()
                            size = int.from_bytes(f.read(4), 'little')
                            value = f.read(size)[:-1].decode("unicode_escape")
                            if id == "ITRK":
                                self.track_number = value
                            if id == "IGNR":
                                self.genre = value
                            if id == "IART":
                                self.artist = value
                            if id == "INAM":
                                self.title = value
                            if id == "IPRD":
                                self.album = value

                            if size % 2 == 1:
                                size += 1
                                f.read(1)

                            remain -= (8 + size)

        wav = wave.open(self.filepath, "rb")
        self.sample_rate = wav.getframerate()
        self.length = wav.getnframes() / self.sample_rate
        wav.close()


genre_dict = {
    0 : 'Blues',
    1 : 'Classic Rock',
    2 : 'Country',
    3 : 'Dance',
    4 : 'Disco',
    5 : 'Funk',
    6 : 'Grunge',
    7 : 'Hip-Hop',
    8 : 'Jazz',
    9 : 'Metal',
    10 : 'New Age',
    11 : 'Oldies',
    12 : 'Other',
    13 : 'Pop',
    14 : 'R&B',
    15 : 'Rap',
    16 : 'Reggae',
    17 : 'Rock',
    18 : 'Techno',
    19 : 'Industrial',
    20 : 'Alternative',
    21 : 'Ska',
    22 : 'Death Metal',
    23 : 'Pranks',
    24 : 'Soundtrack',
    25 : 'Euro-Techno',
    26 : 'Ambient',
    27 : 'Trip-Hop',
    28 : 'Vocal',
    29 : 'Jazz+Funk',
    30 : 'Fusion',
    31 : 'Trance',
    32 : 'Classical',
    33 : 'Instrumental',
    34 : 'Acid',
    35 : 'House',
    36 : 'Game',
    37 : 'Sound Clip',
    38 : 'Gospel',
    39 : 'Noise',
    40 : 'Alternative Rock',
    41 : 'Bass',
    42 : 'Soul',
    43 : 'Punk',
    44 : 'Space',
    45 : 'Meditative',
    46 : 'Instrumental Pop',
    47 : 'Instrumental Rock',
    48 : 'Ethnic',
    49 : 'Gothic',
    50 : 'Darkwave',
    51 : 'Techno-Industrial',
    52 : 'Electronic',
    53 : 'Pop-Folk',
    54 : 'Eurodance',
    55 : 'Dream',
    56 : 'Southern Rock',
    57 : 'Comedy',
    58 : 'Cult',
    59 : 'Gangsta',
    60 : 'Top 40',
    61 : 'Christian Rap',
    62 : 'Pop/Funk',
    63 : 'Jungle',
    64 : 'Native US',
    65 : 'Cabaret',
    66 : 'New Wave',
    67 : 'Psychadelic',
    68 : 'Rave',
    69 : 'Showtunes',
    70 : 'Trailer',
    71 : 'Lo-Fi',
    72 : 'Tribal',
    73 : 'Acid Punk',
    74 : 'Acid Jazz',
    75 : 'Polka',
    76 : 'Retro',
    77 : 'Musical',
    78 : 'Rock & Roll',
    79 : 'Hard Rock',
    80 : 'Folk',
    81 : 'Folk-Rock',
    82 : 'National Folk',
    83 : 'Swing',
    84 : 'Fast Fusion',
    85 : 'Bebob',
    86 : 'Latin',
    87 : 'Revival',
    88 : 'Celtic',
    89 : 'Bluegrass',
    90 : 'Avantgarde',
    91 : 'Gothic Rock',
    92 : 'Progressive Rock',
    93 : 'Psychedelic Rock',
    94 : 'Symphonic Rock',
    95 : 'Slow Rock',
    96 : 'Big Band',
    97 : 'Chorus',
    98 : 'Easy Listening',
    99 : 'Acoustic',
    100 : 'Humour',
    101 : 'Speech',
    102 : 'Chanson',
    103 : 'Opera',
    104 : 'Chamber Music',
    105 : 'Sonata',
    106 : 'Symphony',
    107 : 'Booty Bass',
    108 : 'Primus',
    109 : 'Porn Groove',
    110 : 'Satire',
    111 : 'Slow Jam',
    112 : 'Club',
    113 : 'Tango',
    114 : 'Samba',
    115 : 'Folklore',
    116 : 'Ballad',
    117 : 'Power Ballad',
    118 : 'Rhythmic Soul',
    119 : 'Freestyle',
    120 : 'Duet',
    121 : 'Punk Rock',
    122 : 'Drum Solo',
    123 : 'Acapella',
    124 : 'Euro-House',
    125 : 'Dance Hall',
    126 : 'Goa',
    127 : 'Drum & Bass',
    128 : 'Club - House',
    129 : 'Hardcore',
    130 : 'Terror',
    131 : 'Indie',
    132 : 'BritPop',
    133 : 'Negerpunk',
    134 : 'Polsk Punk',
    135 : 'Beat',
    136 : 'Christian Gangsta Rap',
    137 : 'Heavy Metal',
    138 : 'Black Metal',
    139 : 'Crossover',
    140 : 'Contemporary Christian',
    141 : 'Christian Rock',
    142 : 'Merengue',
    143 : 'Salsa',
    144 : 'Thrash Metal',
    145 : 'Anime',
    146 : 'JPop',
    147 : 'Synthpop',
    148 : 'Unknown',
}


class M4a:

    def __init__(self, file):
        self.filepath = file
        self.has_picture = False

        self.album_artist = ""
        self.artist = ""
        self.genre = ""
        self.date = ""
        self.comment = ""
        self.album = ""
        self.track_number = ""
        self.track_total = ""
        self.title = ""
        self.encoder = ""
        self.disc_number = ""
        self.disc_total = ""
        self.picture = ""
        self.lyrics = ""
        self.track_gain = None
        self.album_gain = None
        self.misc = {}
        self.composer = ""

        self.sample_rate = 0
        self.bit_rate = 0
        self.length = 0

    def read(self, get_picture=False):

        f = open(self.filepath, 'rb')

        k = [
            b"moov",
            b'trak',
            b'----',
            b'udta',
            b'meta',
            b'ilst',
            b'mdia',
            b'mdhd',
            b'minf',
            b'stbl',
            b'stsd',
            b'esds'
        ]

        s_name = b""

        def meta_get(f, size):
            start = f.tell()
            f.seek(16, 1)
            data = f.read(size - 8 - 16)
            f.seek(start)
            return data

        def atom(f, tail=b"", name=""):

            global s_name

            start = f.tell()
            b_size = f.read(4)
            size = int.from_bytes(b_size, 'big')

            name = f.read(4)

            if name == b'':
                return False

            # print("NAME: ", end="")
            # print(tail + b"." + name)

            # Too lazy to parse each sub atom, lets just grab the data out the sub atom and
            # hope the file is formatted normally

            if name == b'\xa9nam':
                self.title = meta_get(f, size).decode().replace("\x00", "")

            if name == b'\xa9alb':
                self.album = meta_get(f, size).decode().replace("\x00", "")

            if name == b'\xa9ART':
                self.artist = meta_get(f, size).decode().replace("\x00", "")

            if name == b'\xa9wrt':
                self.composer = meta_get(f, size).decode().replace("\x00", "")

            if name == b'\xa9cmt':
                self.comment = meta_get(f, size).decode().replace("\x00", "")

            if name == b'\xa9lyr':
                self.lyrics = meta_get(f, size).decode().replace("\x00", "")

            if name == b'\xa9day':
                day = meta_get(f, size).decode().replace("\x00", "")
                if len(day) > 10 and day[10] == "T":
                    day = day[:10]
                self.date = day

            if name == b'gnre':
                index = int.from_bytes(meta_get(f, size), 'big')
                if index - 1 in genre_dict:
                    self.genre = genre_dict[index - 1]

                #self.genre = meta_get(f, size).decode()

            if name == b'\xa9gen':
                self.genre = meta_get(f, size).decode().replace("\x00", "")

            if name == b'aART':
                self.album_artist = meta_get(f, size).decode().replace("\x00", "")

            if name == b'covr':
                self.has_picture = True
                if get_picture:
                    self.picture = meta_get(f, size)

            if name == b'trkn':
                self.track_number = int(meta_get(f, size)[3])  # We'll just grab that

            if name == b'disk':  # They spelt disc wrong lol
                self.disc_number = int(meta_get(f, size)[3])

            # if tail[-4:] == b"----":
            #
            #
            #     if name == b'name':
            #         s_name = f.read(size - 8)
            #         f.seek((size - 8) * -1, 1)
            #
            #     elif name == b'data' and s_name != b"":
            #         data = f.read(size - 8)
            #         f.seek((size - 8) * -1, 1)
            #         print(s_name)
            #         print(data)

            if name in k:

                if name == b'----':
                    f.seek(30, 1)

                if name == b'stsd':
                    f.seek(44, 1)

                if name == b'meta':
                    f.seek(4, 1)  # The 'meta' atom has some extra space at the start

                if name == b'mdhd':

                    data = f.read(size - 8)
                    f.seek((size - 8) * -1, 1)
                    data = struct.unpack('>iiiiihh', data)
                    self.sample_rate = data[3]
                    self.length = data[4] / self.sample_rate

                if name == b'esds':
                    f.seek(26, 1)
                    bitrate = int.from_bytes(f.read(4), 'big')
                    f.seek(-30, 1)
                    self.bit_rate = bitrate // 1000

                else:

                    while f.tell() < start + size:
                        if not atom(f, tail + b"." + name, name):
                            break

            f.seek(start)
            f.seek(size, 1)
            if size == 0:
                return False

            return True

        while atom(f):
            pass



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
import copy
import os
import pickle
import sys
import fcntl
from sdl2 import *
from sdl2.sdlimage import *
from ctypes import pointer
from gi.repository import GLib

# Early arg processing
def transfer_args_and_exit():
    import urllib.request
    base = "http://localhost:7813/"

    if len(sys.argv) <= 1:
        url = base + "raise/"
        urllib.request.urlopen(url)

    for item in sys.argv:

        if not item.endswith(".py") and not item.startswith("-") and (item.startswith("file://") or item.startswith("/")):
            import base64
            url = base + "open/" + base64.urlsafe_b64encode(item.encode()).decode()
            urllib.request.urlopen(url)
        if item == "--play-pause":
            url = base + "playpause/"
            urllib.request.urlopen(url)
        if item == "--play":
            url = base + "play/"
            urllib.request.urlopen(url)
        if item == "--pause":
            url = base + "pause/"
            urllib.request.urlopen(url)
        if item == "--stop":
            url = base + "stop/"
            urllib.request.urlopen(url)
        if item == "--next":
            url = base + "next/"
            urllib.request.urlopen(url)
        if item == "--previous":
            url = base + "previous/"
            urllib.request.urlopen(url)
        if item == "--shuffle":
            url = base + "shuffle/"
            urllib.request.urlopen(url)
        if item == "--repeat":
            url = base + "repeat/"
            urllib.request.urlopen(url)

    sys.exit()


if "--no-start" in sys.argv:
    transfer_args_and_exit()

t_title = 'Tauon Music Box'
os.environ["SDL_VIDEO_X11_WMCLASS"] = t_title  # This sets the window title under some desktop environments

install_directory = os.path.dirname(os.path.abspath(__file__))

pyinstaller_mode = False
if 'base_library' in install_directory:
    pyinstaller_mode = True
    install_directory = os.path.dirname(install_directory)
if hasattr(sys, "_MEIPASS"):
    pyinstaller_mode = True

user_directory = os.path.join(install_directory, "user-data")
config_directory = user_directory

asset_directory = os.path.join(install_directory, "assets")


# If we're installed, use home data locations
install_mode = False
if install_directory.startswith("/opt/")\
        or install_directory.startswith("/usr/")\
        or install_directory.startswith("/app/")\
        or install_directory.startswith("/snap/") or sys.platform == "darwin":
    install_mode = True

if install_mode:
    user_directory = os.path.join(GLib.get_user_data_dir(), "TauonMusicBox")
if not os.path.isdir(user_directory):
    os.makedirs(user_directory)

if os.path.isfile('.gitignore'):
    print("Dev mode, ignoring single instancing")
else:
    pid_file = os.path.join(user_directory, 'program.pid')
    fp = open(pid_file, 'w')
    try:
        fcntl.lockf(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError:
        # another instance is running
        print("Program is already running")
        transfer_args_and_exit()

phone = False
d = os.environ.get('XDG_CURRENT_DESKTOP')
if d in ["GNOME:Phosh"]:
    os.environ["SDL_VIDEODRIVER"] = "wayland"
    phone = True

SDL_SetHint(SDL_HINT_VIDEO_ALLOW_SCREENSAVER, b'1')
SDL_SetHint(SDL_HINT_MOUSE_FOCUS_CLICKTHROUGH, b"1")
SDL_SetHint(SDL_HINT_VIDEO_X11_NET_WM_BYPASS_COMPOSITOR, b"0")

draw_border = True
w = 1120
h = 600
if phone:
    w = 720
    h = 1800
window_default_size = [w, h]
window_size = [w, h]
logical_size = [w, h]
window_opacity = 1
scale = 1
if sys.platform == "darwin":
    scale = 2
if phone:
    scale = 1.3

maximized = False
old_window_position = None

try:
    state_file = open(user_directory + "/window.p", "rb")
    save = pickle.load(state_file)

    draw_border = save[0]
    window_size = save[1]
    logical_size = copy.deepcopy(save[1])
    window_opacity = save[2]
    scale = save[3]
    maximized = save[4]
    old_window_position = save[5]

    del save

except:
    print('No previous window state')

SDL_Init(SDL_INIT_VIDEO | SDL_INIT_GAMECONTROLLER)

err = SDL_GetError()
if err and "GLX" in err.decode():
    print(f"SDL init error: {err.decode()}")
    SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_ERROR, b"Tauon Music Box failed to start :(",
                             b"Error: " + err + b".\n If you're using Flatpak, try run `$ flatpak update`", None)
    sys.exit()

window_title = t_title
window_title = window_title.encode('utf-8')

flags = SDL_WINDOW_RESIZABLE
flags |= SDL_WINDOW_ALLOW_HIGHDPI
if draw_border:
    flags |= SDL_WINDOW_BORDERLESS

if old_window_position is None:
    o_x = SDL_WINDOWPOS_UNDEFINED
    o_y = SDL_WINDOWPOS_UNDEFINED
else:
    o_x = old_window_position[0]
    o_y = old_window_position[1]

if "--tray" in sys.argv:
    flags |= SDL_WINDOW_HIDDEN

t_window = SDL_CreateWindow(window_title,
                            o_x, o_y,
                            logical_size[0], logical_size[1],
                            flags)


if maximized:
    SDL_MaximizeWindow(t_window)

renderer = SDL_CreateRenderer(t_window, 0, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC)
SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_BLEND)
SDL_SetWindowOpacity(t_window, window_opacity)

SDL_SetRenderDrawColor(renderer, 7, 7, 7, 255)
SDL_RenderClear(renderer)


i_x = pointer(c_int(0))
i_y = pointer(c_int(0))
SDL_GL_GetDrawableSize(t_window, i_x, i_y)
window_size[0] = i_x.contents.value
window_size[1] = i_y.contents.value

raw_image = IMG_Load(os.path.join(asset_directory, "loading.png").encode())
sdl_texture = SDL_CreateTextureFromSurface(renderer, raw_image)
w = raw_image.contents.w
h = raw_image.contents.h
rect = SDL_Rect(window_size[0] // 2 - w // 2, window_size[1] // 2 - h // 2, w, h)

SDL_RenderCopy(renderer, sdl_texture, None, rect)

SDL_RenderPresent(renderer)

SDL_FreeSurface(raw_image)
SDL_DestroyTexture(sdl_texture)

from t_modules import t_bootstrap
h = t_bootstrap.holder
h.w = t_window
h.r = renderer
h.wl = logical_size
h.wr = window_size
h.wdf = window_default_size
h.s = scale
h.m = maximized
h.e = transfer_args_and_exit
h.d = draw_border
h.o = window_opacity
h.ow = old_window_position
h.id = install_directory
h.py = pyinstaller_mode
h.p = phone

del raw_image
del sdl_texture
del w
del h
del rect
del flags

if pyinstaller_mode or sys.platform == "darwin":
    from t_modules import t_main
else:
    # Using the above import method breaks previous pickles. Could be fixed
    # but yet to decide what best method is.
    big_boy_path = os.path.join(install_directory, 't_modules/t_main.py')
    f = open(big_boy_path, "rb")
    main = compile(f.read(), big_boy_path, 'exec')
    f.close()
    del big_boy_path
    del f
    exec(main)





# Tauon Music Box - Misc Functions Module

# Copyright © 2015-2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.


import time
import random
import colorsys
import subprocess
import os
import shlex
import zipfile
import glob
import locale
import re
import math
import threading
import urllib.parse

# A seconds based timer
class Timer:
    def __init__(self, force=None):
        self.start = 0
        self.end = 0
        self.set()
        if force:
            self.force_set(force)

    def set(self):  # Reset
        self.start = time.monotonic()

    def hit(self):  # Return time and reset

        self.end = time.monotonic()
        elapsed = self.end - self.start
        self.start = time.monotonic()
        return elapsed

    def get(self):  # Return time only
        self.end = time.monotonic()
        return self.end - self.start

    def force_set(self, sec):
        self.start = time.monotonic()
        self.start -= sec


# Simple bool timer object
class TestTimer:
    def __init__(self, time):
        self.timer = Timer()
        self.time = time

    def test(self):
        return self.timer.get() > self.time



# Test given proximity between two 2d points to given square
def point_proximity_test(a, b, p):
    return abs(a[0] - b[0]) < p and abs(a[1] - b[1]) < p

# Get distance between two points
def point_distance(a, b):
    return math.sqrt(abs(a[0] - b[0]) ** 2 + abs(b[1] - b[1]) ** 2)

# Removes whatever this is from a line, I forgot
def rm_16(line):
    if "ÿ þ" in line:
        for c in line:
            line = line[1:]
            if c == 'þ':
                break

        line = line[::2]
    return line


# Returns a string from seconds to a compact time format, e.g 2h:23
def get_display_time(seconds):
    result = divmod(int(seconds), 60)
    if result[0] > 99:
        result = divmod(result[0], 60)
        return str(result[0]) + 'h ' + str(result[1]).zfill(2)
    return str(result[0]).zfill(2) + ":" + str(result[1]).zfill(2)


def get_hms_time(seconds):
    m, s = divmod(round(seconds), 60)
    h, m = divmod(m, 60)
    if h:
        return f'{h:d}:{m:02d}:{s:02d}'
    else:
        return f'{m:02d}:{s:02d}'


# Creates a string from number of bytes to X MB/kB etc with Locale adjustment
def get_filesize_string(file_bytes):
    if not file_bytes:
        return "0"
    if file_bytes < 1000:
        line = locale.str(file_bytes) + " B"
    elif file_bytes < 1000000:
        file_kb = round(file_bytes / 1000, 2)
        line = locale.str(file_kb) + " KB"
    elif file_bytes < 1000000000:
        file_mb = round(file_bytes / 1000000, 2)
        line = locale.str(file_mb) + " MB"
    else:
        file_mb = round(file_bytes / 1000000000, 1)
        line = locale.str(file_mb) + " GB"
    return line

def get_filesize_string_rounded(file_bytes):
    if not file_bytes:
        return "0"
    if file_bytes < 1000:
        line = str(round(file_bytes)) + " B"
    elif file_bytes < 1000000:
        file_kb = round(file_bytes / 1000)
        line = str(file_kb) + " KB"
    else:
        file_mb = round(file_bytes / 1000000, 1)
        line = str(file_mb) + " MB"
    return line

# Estimates the perceived luminance of a colour
def test_lumi(c1):
    return 1 - (0.299 * c1[0] + 0.587 * c1[1] + 0.114 * c1[2]) / 255

def rel_luminance(colour):
    r = colour[0] / 255
    g = colour[1] / 255
    b = colour[2] / 255

    if r < 0.03928:
        r = r / 12.90
    else:
        r = ((r + 0.055) / 1.055) ** 2.4

    if g < 0.03928:
        g = g / 12.90
    else:
        g = ((g + 0.055) / 1.055) ** 2.4

    if b < 0.03928:
        b = b / 12.90
    else:
        b = ((b + 0.055) / 1.055) ** 2.4

    return 0.2126 * r + 0.7152 * g + 0.0722 * b


def contrast_ratio(c1, c2):

    l1 = rel_luminance(c1)
    l2 = rel_luminance(c2)

    if l2 > l1:
        l2, l1 = l1, l2

    return (l1 + 0.05) / (l2 + 0.05)

# Gives the sum of first 3 elements in a list
def colour_value(c1):
    return c1[0] + c1[1] + c1[2]


# Performs alpha blending of one colour (rgba) onto another (rgb)
def alpha_blend(colour, base):
    alpha = colour[3] / 255
    return [int(alpha * colour[0] + (1 - alpha) * base[0]),
            int(alpha * colour[1] + (1 - alpha) * base[1]),
            int(alpha * colour[2] + (1 - alpha) * base[2]), 255]


# Change the alpha component of an RGBA list
def alpha_mod(colour, alpha):
    return [colour[0], colour[1], colour[2], alpha]


# Shift between two colours based on x where x is between 0 and limit
def colour_slide(a, b, x, x_limit):

    return (min(int(a[0] + ((b[0] - a[0]) * (x / x_limit))), 255),
     min(int(a[1] + ((b[1] - a[1]) * (x / x_limit))), 255),
     min(int(a[2] + ((b[2] - a[2]) * (x / x_limit))), 255), 255)


def hex_to_rgb(colour):
    colour = colour.strip("#")
    return list(int(colour[i:i + 2], 16) for i in (0, 2, 4)) + [255]


# Checks if all the numbers in a list are the same
def check_equal(lst):
    return not lst or lst.count(lst[0]) == len(lst)


# Check if the first 3 elements of a list are the same
def is_grey(lst):
    return lst[0] == lst[1] == lst[2]


# Gives a score from 0-7 based on number of seconds
def star_count(sec, dur):
    stars = 0
    if dur and sec / dur > 0.95:
        stars += 1
    if sec > 60 * 15:
        stars += 1
    if sec > 60 * 30:
        stars += 1
    if sec > 60 * 60:
        stars += 1
    if sec > 60 * 60 * 2:
        stars += 1
    if sec > 60 * 60 * 4:
        stars += 1
    if sec > 60 * 60 * 8:
        stars += 1
    if sec > 60 * 60 * 16:
        stars += 1
    return stars


def star_count3(sec, dur):
    stars = 0
    if dur and sec / dur > 0.95:
        stars += 1
    if dur and sec / dur > 1.95:
        stars += 1
    if sec > 60 * 5:
        stars += 1
    if sec > 60 * 10:
        stars += 1
    if sec > 60 * 15:
        stars += 1
    if sec > 60 * 20:
        stars += 1
    if sec > 60 * 30:
        stars += 1
    if sec > 60 * 60:
        stars += 1
    if sec > 60 * 60 * 1.5:
        stars += 1
    if sec > 60 * 60 * 1.75:
        stars += 1
    # if sec > 60 * 60 * 2:
    #     stars += 1
    return stars


# Gives a score from 0.0 - 1.0 based on number of seconds
def star_count2(sec):
    star = 0
    star += min(sec / (60 * 4), 0.2)
    star += min(sec / (60 * 60), 0.4)
    star += sec / (60 * 60 * 10)
    return float(round(min(star, 1), 1))


def search_magic(terms, evaluate):
    return all(word in evaluate for word in terms.split())


def search_magic_any(terms, evaluate):
    return any(word in evaluate for word in terms.split())


def random_colour(saturation, luminance):

    h = round(random.random(), 2)
    colour = colorsys.hls_to_rgb(h, luminance, saturation)
    return [int(colour[0] * 255), int(colour[1] * 255), int(colour[2] * 255), 255]


def hsl_to_rgb(h, s, l):
    colour = colorsys.hls_to_rgb(h, l, s)
    return [int(colour[0] * 255), int(colour[1] * 255), int(colour[2] * 255), 255]

def hls_to_rgb(h, l, s):
    colour = colorsys.hls_to_rgb(h, l, s)
    return [int(colour[0] * 255), int(colour[1] * 255), int(colour[2] * 255), 255]

def rgb_to_hls(r, g, b):
    return colorsys.rgb_to_hls(r / 255, g / 255, b / 255)

def rgb_add_hls(source, h=0, l=0, s=0):
    c = colorsys.rgb_to_hls(source[0] / 255, source[1] / 255, source[2] / 255)
    colour = colorsys.hls_to_rgb(c[0] + h, min(max(c[1] + l, 0), 1), min(max(c[2] + s, 0), 1))
    return [int(colour[0] * 255), int(colour[1] * 255), int(colour[2] * 255), source[3]]


def is_light(colour):
    return test_lumi(colour) < 0.2

class ColourGenCache:

    def __init__(self, saturation, luminance):

        self.saturation = saturation
        self.luminance = luminance
        self.store = {}

    def get(self, key):

        if key in self.store:
            return self.store[key]

        colour = random_colour(self.saturation, self.luminance)

        self.store[key] = colour
        return colour




def folder_file_scan(path, extensions):

    match = 0
    count = sum([len(files) for r, d, files in os.walk(path)])
    for ext in extensions:

        match += len(glob.glob(path + '/**/*.' + ext.lower(), recursive=True))

    if count == 0:
        return 0

    if count < 5 and match > 0:
        return 1

    return match / count


def is_ignorable_file(string):
    for s in [
        "Thumbs.db",
        ".log",
        "desktop.ini",
        "DS_Store",
        ".nfo",
        "yric"
    ]:
        if s in string:
            return True
    return False


def is_music_related(string):
    for s in [
        "Folder.jpg",
        "folder.jpg",
        "Cover.jpg",
        "cover.jpg",
        "AlbumArt",
        ".m3u",
        ".m3u8",
        ".cue",
        ".CUE",
    ]:
        if s in string:
            return True
    return False


# Get ratio of given file extensions in archive
def archive_file_scan(path, extensions, launch_prefix=""):

    ext = os.path.splitext(path)[1][1:].lower()
    # print(path)
    # print(ext)
    try:
        if ext == 'rar':
            matches = 0
            count = 0
            line = launch_prefix + "unrar lb -p- " + shlex.quote(path) + " " + shlex.quote(os.path.dirname(path)) + os.sep
            result = subprocess.run(shlex.split(line), stdout=subprocess.PIPE)
            file_list = result.stdout.decode("utf-8", 'ignore').split("\n")
            # print(file_list)
            for fi in file_list:
                for ty in extensions:
                    if fi[len(ty) * -1:].lower() == ty:
                        matches += 1
                        break
                    elif is_ignorable_file(fi):
                        count -= 1
                        break
                    elif is_music_related(fi):
                        matches += 5
                count += 1
            if count > 200:
                #print("RAR archive has many files")
                #print("   --- " + path)
                return 0
            if matches == 0:
                #print("RAR archive does not appear to contain audio files")
                #print("   --- " + path)
                return 0
            if count == 0:
                #print("Archive has no files")
                #print("   --- " + path)
                return 0

        elif ext == '7z':
            matches = 0
            count = 0
            line = launch_prefix + "7z l " + shlex.quote(path) # + " " + shlex.quote(os.path.dirname(path)) + os.sep
            result = subprocess.run(shlex.split(line), stdout=subprocess.PIPE)
            file_list = result.stdout.decode("utf-8", 'ignore').split("\n")
            # print(file_list)

            for fi in file_list:

                if '....A' not in fi:
                    continue
                for ty in extensions:
                    if fi[len(ty) * -1:].lower() == ty:
                        matches += 1
                        break
                    elif is_ignorable_file(fi):
                        count -= 1
                        break
                    elif is_music_related(fi):
                        matches += 5
                count += 1

            if count > 200:
                #print("7z archive has many files")
                #print("   --- " + path)
                return 0
            if matches == 0:
                #print("7z archive does not appear to contain audio files")
                #print("   --- " + path)
                return 0
            if count == 0:
                #print("7z archive has no files")
                #print("   --- " + path)
                return 0

        elif ext == "zip":

            zip_ref = zipfile.ZipFile(path, 'r')
            matches = 0
            count = 0
            #print(zip_ref.namelist())
            for fi in zip_ref.namelist():
                for ty in extensions:
                    if fi[len(ty) * -1:].lower() == ty:
                        matches += 1
                        break
                    elif is_ignorable_file(fi):
                        count -= 1
                        break
                    elif is_music_related(fi):
                        matches += 5
                count += 1
            if count == 0:
                #print("Archive has no files")
                #print("   --- " + path)
                return 0
            if count > 300:
                #print("Zip archive has many files")
                #print("   --- " + path)
                return 0
            if matches == 0:
                #print("Zip archive does not appear to contain audio files")
                #print("   --- " + path)
                return 0
        else:
            return 0

    except:
        print("Archive test error")

        return 0

    if count == 0:
        return 0

    ratio = matches / count
    if count < 5 and matches > 0:
        ratio = 100
    return ratio


def get_folder_size(path):
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += os.path.getsize(fp)
    return total_size


def filename_safe(text, sub=""):
    for cha in '/\\<>:"|?*':
        text = text.replace(cha, sub)
    return text

def get_artist_strip_feat(track_object):

    artist_name = track_object.artist #.lower()
    if track_object.album_artist:
        if "feat." in artist_name or "pres." in artist_name or ", " in artist_name or "; " in artist_name or not artist_name:
            if track_object.album_artist.lower() != "va" and \
                    track_object.album_artist.lower() != "various artists":
                artist_name = track_object.album_artist
    return artist_name

def get_artist_safe(track):

    if track:
        artist = track.album_artist
        if not artist:
            artist = track.artist
        artist = filename_safe(artist)
        artist = artist.split("feat")[0]
        artist = artist.split(", ")[0]
        artist = artist.split("; ")[0]
        return artist
    return ""

def get_split_artists(track):
    if 'artists' in track.misc:
        return track.misc['artists']
    artist = track.artist.split("feat")[0].strip()
    return re.split(r'; |, |& ', artist)

def coll_rect(rect1, rect2):

    if rect1[0] + rect1[2] < rect2[0] or \
            rect1[1] + rect1[3] < rect2[1] or \
            rect1[0] > rect2[0] + rect2[2] or \
            rect1[1] > rect2[1] + rect2[3]:
        return False
    return True


def commonprefix(l):

    cp = []
    ls = [p.split('/') for p in l]
    ml = min(len(p) for p in ls)

    for i in range(ml):

        s = set(p[i] for p in ls)
        if len(s) != 1:
            break

        cp.append(s.pop())

    return '/'.join(cp)


def fader_timer(time_point, start, duration, off=True, range=255):

    if time_point < start:
        fade = range
    elif time_point < start + duration:
        p = (time_point - start) / duration
        fade = int(range - (range * p))
    else:
        fade = 0

    return fade


id3_genre_dict = {
    
    0: "Blues",
    1: "Classic Rock",
    2: "Country",
    3: "Dance",
    4: "Disco",
    5: "Funk",
    6: "Grunge",
    7: "Hip-Hop",
    8: "Jazz",
    9: "Metal",
    10: "New Age",
    11: "Oldies",
    12: "Other",
    13: "Pop",
    14: "R&B",
    15: "Rap",
    16: "Reggae",
    17: "Rock",
    18: "Techno",
    19: "Industrial",
    20: "Alternative",
    21: "Ska",
    22: "Death Metal",
    23: "Pranks",
    24: "Soundtrack",
    25: "Euro-Techno",
    26: "Ambient",
    27: "Trip-Hop",
    28: "Vocal",
    29: "Jazz+Funk",
    30: "Fusion",
    31: "Trance",
    32: "Classical",
    33: "Instrumental",
    34: "Acid",
    35: "House",
    36: "Game",
    37: "Sound Clip",
    38: "Gospel",
    39: "Noise",
    40: "Alternative Rock",
    41: "Bass",
    42: "Soul",
    43: "Punk",
    44: "Space",
    45: "Meditative",
    46: "Instrumental Pop",
    47: "Instrumental Rock",
    48: "Ethnic",
    49: "Gothic",
    50: "Darkwave",
    51: "Techno-Industrial",
    52: "Electronic",
    53: "Pop-Folk",
    54: "Eurodance",
    55: "Dream",
    56: "Southern Rock",
    57: "Comedy",
    58: "Cult",
    59: "Gangsta Rap",
    60: "Top 40",
    61: "Christian Rap",
    62: "Pop/Funk",
    63: "Jungle",
    64: "Native American",
    65: "Cabaret",
    66: "New Wave",
    67: "Psychedelic",
    68: "Rave",
    69: "Showtunes",
    70: "Trailer",
    71: "Lo-Fi",
    72: "Tribal",
    73: "Acid Punk",
    74: "Acid Jazz",
    75: "Polka",
    76: "Retro",
    77: "Musical",
    78: "Rock & Roll",
    79: "Hard Rock",
    80: "Folk",
    81: "Folk-Rock",
    82: "National Folk",
    83: "Swing",
    84: "Fast Fusion",
    85: "Bebob",
    86: "Latin",
    87: "Revival",
    88: "Celtic",
    89: "Bluegrass",
    90: "Avantgarde",
    91: "Gothic Rock",
    92: "Progressive Rock",
    93: "Psychedelic Rock",
    94: "Symphonic Rock",
    95: "Slow Rock",
    96: "Big Band",
    97: "Chorus",
    98: "Easy Listening",
    99: "Acoustic",
    100: "Humour",
    101: "Speech",
    102: "Chanson",
    103: "Opera",
    104: "Chamber Music",
    105: "Sonata",
    106: "Symphony",
    107: "Booty Bass",
    108: "Primus",
    109: "Porn Groove",
    110: "Satire",
    111: "Slow Jam",
    112: "Club",
    113: "Tango",
    114: "Samba",
    115: "Folklore",
    116: "Ballad",
    117: "Power Ballad",
    118: "Rhythmic Soul",
    119: "Freestyle",
    120: "Duet",
    121: "Punk Rock",
    122: "Drum Solo",
    123: "A Cappella",
    124: "Euro-House",
    125: "Dance Hall",
    126: "Goa",
    127: "Drum & Bass",
    128: "Club-House",
    129: "Hardcore",
    130: "Terror",
    131: "Indie",
    132: "BritPop",
    133: "Negerpunk",
    134: "Polsk Punk",
    135: "Beat",
    136: "Christian Gangsta Rap",
    137: "Heavy Metal",
    138: "Black Metal",
    139: "Crossover",
    140: "Contemporary Christian",
    141: "Christian Rock",
    142: "Merengue",
    143: "Salsa",
    144: "Thrash Metal",
    145: "Anime",
    146: "JPop",
    147: "Synthpop",
    148: "Abstract",
    149: "Art Rock",
    150: "Baroque",
    151: "Bhangra",
    152: "Big Beat",
    153: "Breakbeat",
    154: "Chillout",
    155: "Downtempo",
    156: "Dub",
    157: "EBM",
    158: "Eclectic",
    159: "Electro",
    160: "Electroclash",
    161: "Emo",
    162: "Experimental",
    163: "Garage",
    164: "Global",
    165: "IDM",
    166: "Illbient",
    167: "Industro-Goth",
    168: "Jam Band",
    169: "Krautrock",
    170: "Leftfield",
    171: "Lounge",
    172: "Math Rock",
    173: "New Romantic",
    174: "Nu-Breakz",
    175: "Post-Punk",
    176: "Post-Rock",
    177: "Psytrance",
    178: "Shoegaze",
    179: "Space Rock",
    180: "Trop Rock",
    181: "World Music",
    182: "Neoclassical",
    183: "Audiobook",
    184: "Audio Theatre",
    185: "Neue Deutsche Welle",
    186: "Podcast",
    187: "Indie Rock",
    188: "G-Funk",
    189: "Dubstep",
    190: "Garage Rock",
    191: "Psybient",
    192: "Unknown",
}


class FunctionStore:
    # Stores functions and arguments for calling later

    def __init__(self):

        self.items = []

    def store(self, function, args=()):

        self.items.append((function, args))

    def recall_all(self):

        while self.items:
            item = self.items.pop()
            item[0](*item[1])


year_search = re.compile(r'\d{4}')



def year_from_string(date_string):

    result = year_search.search(date_string)
    if result:
        return result.group(0)
    else:
        return False


def grow_rect(rect, px):
    return rect[0] - px, rect[1] - px, rect[2] + px * 2, rect[3] + px * 2

# def get_hash(f_path, mode='sha256'):
#     h = hashlib.new(mode)
#     with open(f_path, 'rb') as file:
#         data = file.read()
#     h.update(data)
#     digest = h.hexdigest()
#     return digest

def subtract_rect(base, hole):  # Return 4 rects from 1 minus 1 inner (with overlaps)

    west = base[0], base[1], hole[0], base[3]
    north = base[0], base[1], base[2], hole[1] - base[1]
    east = base[0] + hole[0] + hole[2], base[1], base[2] - (hole[0] + hole[2]), base[3]
    south = base[0], hole[1] + hole[3], base[2], base[3] - hole[3] - 2

    return west, north, east, south

genre_corrections = [
    "J-Pop",
    "J-Rock",
    "K-Pop",
    "Hip Hop",
]

genre_corrections2 = [x.lower().replace("-", "").replace(" ", "") for x in genre_corrections]

def genre_correct(text):

    parsed = text.lower().replace("-", "").replace(" ", "").strip()
    if parsed.startswith("post"):
        return ("Post-" + parsed[4:]).title()
    if parsed in genre_corrections2:
        return genre_corrections[genre_corrections2.index(parsed)]
    return text.title().strip()


def reduce_paths(paths):  # in-place remove of redundant sub-paths from list of folder paths

    paths[:] = list(set(paths))[:]  # remove duplicates

    while "" in paths:
        paths.remove("")

    while True:
        remove_path = False
        for i in reversed(range(len(paths))):
            path = paths[i].rstrip("/")

            for b in reversed(range(len(paths))):
                path2 = paths[b].rstrip("/")

                if len(path) > len(path2) and path.startswith(path2) and path[len(path2)] == "/":
                    del paths[i]
                    remove_path = True
                    break

            if remove_path:
                break
        if not remove_path:
            break

def fit_box(inner, outer):
    scale = min(outer[0]/inner[0], outer[1]/inner[1])
    return round(inner[0] * scale), round(inner[1] * scale)


def seconds_to_day_hms(seconds, s_day, s_days):

    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minuites, seconds = divmod(seconds, 60)

    if days == 1:
        return f"{str(int(days))} {s_day}, {str(int(hours))}:{str(int(minuites))}:{str(int(seconds))}"
    else:
        return f"{str(int(days))} {s_days}, {str(int(hours))}:{str(int(minuites)).zfill(2)}:{str(int(seconds)).zfill(2)}"


def shooter(func, args=()):
    shoot = threading.Thread(target=func, args=args)
    shoot.daemon = True
    shoot.start()

def d_date_display(track):
    if 'rdat' in track.misc:
        return str(track.date) + " → " + track.misc["rdat"]
    return str(track.date)

def process_odat(nt, odat):
    if odat and odat != nt.date and odat != nt.date[:4] and odat != nt.date[-4:] \
            and nt.date != odat[:4] and nt.date != odat[-4:]:
        if not nt.date:
            nt.date = odat
        else:
            nt.misc["rdat"] = nt.date
            nt.date = odat

def clean_string(s):
    return s.encode('utf-8', 'surrogatepass').decode('utf-8', 'replace')

def uri_parse(s):
    if s.startswith("file://"):
        s = str(urllib.parse.unquote(s)).replace("file://", "").replace("\r", "")
    return s

mac_styles = {
    "mac": None,
    "whitesur": None,
    "vimix": None,
    "sweet": None,
    "dracula": [[248, 58, 67, 255], [239, 251, 122, 255], [74, 254, 104, 255]],
    "nordic": None,
    "juno": None
}



# Tauon Music Box - Jellyin client API module

# Copyright © 2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.

import requests
import json
import itertools
import io
import time
import threading
from t_modules.t_extra import Timer

class Jellyfin():

    def __init__(self, tauon):
        self.tauon = tauon
        self.pctl = tauon.pctl
        self.prefs = tauon.prefs
        self.gui = tauon.gui

        self.scanning = False
        self.connected = False

        self.accessToken = None
        self.userId = None
        self.currentId = None

        self.session_thread_active = False
        self.session_status = 0
        self.session_item_id = None
        self.session_update_timer = Timer()
        self.session_last_item = None


    def _get_jellyfin_auth(self):
        auth_str = f"MediaBrowser Client={self.tauon.t_title}, Device={self.tauon.device}, DeviceId=-, Version={self.tauon.t_version}"
        if self.accessToken:
            auth_str += f", Token={self.accessToken}"
        return auth_str

    def _authenticate(self, debug=False):
        username = self.prefs.jelly_username
        password = self.prefs.jelly_password
        server = self.prefs.jelly_server_url

        try:
            response = requests.post(
                f"{server}/Users/AuthenticateByName",
                headers={
                    "Content-type": "application/json",
                    "X-Application": self.tauon.t_agent,
                    "x-emby-authorization": self._get_jellyfin_auth()
                },
                data=json.dumps({ "username": username, "Pw": password }),
            )
        except:
            self.gui.show_message("Could not establish connection to server.", "Check server is running and URL is correct.", mode="error")
            return

        if response.status_code == 200:
            info = response.json()
            self.accessToken = info["AccessToken"]
            self.userId = info["User"]["Id"]
            self.connected = True
            if debug:
                self.gui.show_message("Connection and authorisation successful", mode="done")
        else:
            if response.status_code == 401:
                self.gui.show_message("401 Authentication failed", "Check username and password.", mode="warning")
            else:
                self.gui.show_message("Jellyfin auth error", f"{response.status_code} {response.text}", mode="warning")

    def test(self):
        self._authenticate(debug=True)

    def resolve_stream(self, id):
        if not self.connected or not self.accessToken:
            self._authenticate()

        if not self.connected:
            return ""

        base_url = f"{self.prefs.jelly_server_url}/Audio/{id}/stream"
        headers = {
            "Token": self.accessToken,
            "X-Application": "Tauon/1.0",
            "x-emby-authorization": self._get_jellyfin_auth()
        }
        params = {
            "UserId": self.userId,
            "static": "true"
        }

        if self.prefs.network_stream_bitrate > 0:
            params["MaxStreamingBitrate"] = self.prefs.network_stream_bitrate

        url = base_url + "?" + requests.compat.urlencode(params)

        if not self.session_thread_active:
            shoot = threading.Thread(target=self.session)
            shoot.daemon = True
            shoot.start()

        return base_url, params

    def get_cover(self, track):
        if not self.connected or not self.accessToken:
            self._authenticate()

        if not self.connected:
            return None

        if not track.art_url_key:
            return None

        headers = {
            "Token": self.accessToken,
            "X-Application": "Tauon/1.0",
            "x-emby-authorization": self._get_jellyfin_auth()
        }
        params = {}
        base_url = f"{self.prefs.jelly_server_url}/Items/{track.art_url_key}/Images/Primary"
        response = requests.get(base_url, headers=headers, params=params)

        if response.status_code == 200:
            return io.BytesIO(response.content)
        else:
            print("Jellyfin album art api error:", response.status_code, response.text)
            return None

    def favorite(self, track, un=False):
        try:
            if not self.connected or not self.accessToken:
                self._authenticate()

            if not self.connected:
                return None

            headers = {
                "Token": self.accessToken,
                "X-Application": "Tauon/1.0",
                "x-emby-authorization": self._get_jellyfin_auth()
            }

            params = {}
            base_url = f"{self.prefs.jelly_server_url}/Users/{self.userId}/FavoriteItems/{track.url_key}"
            if un:
                response = requests.delete(base_url, headers=headers, params=params)
            else:
                response = requests.post(base_url, headers=headers, params=params)

            if response.status_code == 200:
                return
            else:
                print("Jellyfin fav api error")

        except:
            print("Failed to submit favorite to Jellyfin server")

    def ingest_library(self, return_list=False):
        self.gui.update += 1
        self.scanning = True
        self.gui.to_got = 0

        print("Prepare for Jellyfin library import...")

        if not self.connected or not self.accessToken:
            self._authenticate()

        if not self.connected:
            self.scanning = False
            if not return_list:
                self.tauon.gui.show_message("Error connecting to Jellyfin")
            return []

        playlist = []

        # This code is to identify if a track has already been imported
        existing = {}
        for track_id, track in self.pctl.master_library.items():
            if track.is_network and track.file_ext == "JELY":
                existing[track.url_key] = track_id

        print("Get items...")

        try:
            response = requests.get(
                f"{self.prefs.jelly_server_url}/Users/{self.userId}/Items",
                headers={
                    "Token": self.accessToken,
                    "X-Application": "Tauon/1.0",
                    "x-emby-authorization": self._get_jellyfin_auth()
                },
                params={
                    "recursive": True,
                    "filter": "music"
                }
            )
        except:
            self.gui.show_message("Error connecting to Jellyfin for Import", mode="error")
            self.scanning = False
            return

        if response.status_code == 200:

            print("Connection successful, soring items...")

            # filter audio items only
            audio_items = list(filter(lambda item: item["Type"] == "Audio", response.json()["Items"]))
            # sort by artist, then album, then track number
            sorted_items = sorted(audio_items, key=lambda item: (item.get("AlbumArtist", ""), item.get("Album", ""), item.get("IndexNumber", -1)))
            # group by parent
            grouped_items = itertools.groupby(sorted_items, lambda item: (item.get("AlbumArtist", "") + " - " + item.get("Album", "")).strip("- "))
        else:
            self.scanning = False
            self.tauon.gui.show_message("Error accessing Jellyfin", mode="warning")
            return

        mem_folder = {}
        for parent, items in grouped_items:
            for track in items:
                id = self.pctl.master_count  # id here is tauons track_id for the track
                existing_track = existing.get(track.get("Id"))
                replace_existing = existing_track is not None
                #print(track.items())
                if replace_existing:
                    id = existing_track

                nt = self.tauon.TrackClass()
                nt.index = id  # this is tauons track id
                nt.track_number = str(track.get("IndexNumber", ""))
                nt.disc_number = str(track.get("ParentIndexNumber", ""))
                nt.file_ext = "JELY"
                nt.album_artist = track.get("AlbumArtist", "")
                artists = track.get("Artists", [])
                nt.artist = "; ".join(artists)
                if len(artists) > 1:
                    nt.misc["artists"] = artists
                nt.title = track.get("Name", "")
                nt.album = track.get("Album", "")
                nt.length = track.get("RunTimeTicks", 0) / 10000000   # needs to be in seconds
                nt.date = str(track.get("ProductionYear"))

                folder_name = nt.album_artist
                if folder_name and nt.album:
                    folder_name += " / "
                folder_name += nt.album

                if track.get("AlbumId") and folder_name:
                    key = track.get("AlbumId") + nt.album
                    if key not in mem_folder:
                        mem_folder[key] = folder_name
                    folder_name = mem_folder[key]
                elif nt.album:
                    folder_name = nt.album

                nt.parent_folder_path = folder_name
                nt.parent_folder_name = nt.parent_folder_path
                nt.is_network = True

                nt.url_key = track.get("Id")
                nt.art_url_key = track.get("AlbumId") if track.get("AlbumPrimaryImageTag", False) else None

                self.pctl.master_library[id] = nt
                if not replace_existing:
                    self.pctl.master_count += 1
                playlist.append(nt.index)

                # Sync favorite
                star = self.tauon.star_store.full_get(nt.index)
                user_data = track.get("UserData")
                if user_data:
                    if user_data.get("IsFavorite"):
                        if star is None:
                            star = self.tauon.star_store.new_object()
                        if 'L' not in star[1]:
                            star[1] += "L"
                        self.tauon.star_store.insert(nt.index, star)
                    else:
                        if star is None:
                            pass
                        else:
                            star = [star[0], star[1].replace("L", ""), star[2]]
                            self.tauon.star_store.insert(nt.index, star)

        self.scanning = False
        print("Jellyfin import complete")

        playlist.sort(key=lambda x: self.pctl.master_library[x].parent_folder_path)
        self.tauon.sort_track_2(0, playlist)

        if return_list:
            return playlist

        self.pctl.multi_playlist.append(self.tauon.pl_gen(title="Jellyfin Collection", playlist=playlist))
        self.pctl.gen_codes[self.tauon.pl_to_id(len(self.pctl.multi_playlist) - 1)] = "jelly"
        self.tauon.switch_playlist(len(self.pctl.multi_playlist) - 1)

    def session_item(self, track):
        return {
            "QueueableMediaTypes": ["Audio"],
            "CanSeek": True,
            "ItemId": track.url_key,
            "IsPaused": self.pctl.playing_state == 2,
            "IsMuted": self.pctl.player_volume == 0,
            "PositionTicks": int(self.pctl.playing_time * 10000000),
            "PlayMethod": "DirectStream",
            "PlaySessionId": "0",
        }

    def session(self):

        if not self.connected:
            return

        self.session_thread_active = True

        while True:
            time.sleep(1)
            track = self.pctl.playing_object()

            if track.file_ext != "JELY" or (self.session_status == 0 and self.pctl.playing_state == 0):
                if self.session_status != 0:
                    data = self.session_last_item
                    self.session_send("Sessions/Playing/Stopped", data)
                    self.session_status = 0
                self.session_thread_active = False
                return

            if (self.session_status == 0 or self.session_item_id != track.index) and self.pctl.playing_state == 1:
                data = self.session_item(track)
                self.session_send("Sessions/Playing", data)
                self.session_update_timer.set()
                self.session_status = 1
                self.session_item_id = track.index
                self.session_last_item = data
            elif self.session_status == 1 and self.session_update_timer.get() >= 10:
                data = self.session_item(track)
                data["EventName"] = "TimeUpdate"
                self.session_send("Sessions/Playing/Progress", data)
                self.session_update_timer.set()
            elif self.session_status in (1, 2) and self.pctl.playing_state in (0, 3):
                data = self.session_last_item
                self.session_send("Sessions/Playing/Stopped", data)
                self.session_status = 0
            elif self.session_status == 1 and self.pctl.playing_state == 2:
                data = self.session_item(track)
                data["EventName"] = "Pause"
                self.session_send("Sessions/Playing/Progress", data)
                self.session_update_timer.set()
                self.session_status = 2
            elif self.session_status == 2 and self.pctl.playing_state == 1:
                data = self.session_item(track)
                data["EventName"] = "Unpause"
                self.session_send("Sessions/Playing/Progress", data)
                self.session_update_timer.set()
                self.session_status = 1

    def session_send(self, point, data):

        response = requests.post(
            f"{self.prefs.jelly_server_url}/{point}", data=json.dumps(data),
            headers={
                "Token": self.accessToken,
                "X-Application": "Tauon/1.0",
                "x-emby-authorization": self._get_jellyfin_auth(),
                "Content-Type": "application/json"
            })



# Tauon Music Box - Lyrics scrape/fetch module

# Copyright © 2018-2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.


from isounidecode import unidecode
from bs4 import BeautifulSoup  # Remember to add to dependency list if PyLyrics is removed
import urllib.parse
import requests  # Remember to add to dependency list if PyLyrics is removed
import re


# You can add lyric providers in this module

# Create a function that takes artist and title, and returns lyrics as a str
# If failed to find lyrics, you can return None or raise an exception.
# Finally add provider name and function reference to lyric_sources dict below


# def lyricwiki(artist, title):
#
#     lyrics = PyLyrics.getLyrics(artist, title)
#
#     if lyrics and lyrics[0] == "<" and "Instrumental" in lyrics:
#         lyrics = "[Instrumental]"
#
#     return lyrics

def ovh(artist, title):

    q = urllib.parse.quote(f"{artist}/{title}")
    point = f"https://api.lyrics.ovh/v1/{q}"
    r = requests.get(point)
    if r.status_code == 200:
        j = r.json()
        return j["lyrics"]
    return ""


def happi(artist, title):
    q = urllib.parse.quote(f"{artist} {title}")
    point = f"https://api.happi.dev/v1/music?q={q}&limit=1&apikey=23b23b30Ca5nqZSe5JWJ8I4smmgO1JK6grVTEXpkBz1O8mNjTCmmCjnX&type=track"
    r = requests.get(point)
    j = r.json()
    if not j["result"][0]["haslyrics"]:
        return ""
    a_id = j["result"][0]["id_artist"]
    t_id = j["result"][0]["id_track"]
    al_id = j["result"][0]["id_album"]

    point = f"https://api.happi.dev/v1/music/artists/{a_id}/albums/{al_id}/tracks/{t_id}/lyrics?apikey=23b23b30Ca5nqZSe5JWJ8I4smmgO1JK6grVTEXpkBz1O8mNjTCmmCjnX"
    r = requests.get(point)
    j = r.json()
    return j["result"]["lyrics"]

def genius(artist, title, return_url=False):

    line = f"{artist}-{title}"
    line = re.sub("[,._@!#%^*+:;'()]", "", line)
    line = line.replace("]", "")
    line = line.replace("[", "")
    line = line.replace("?", "")
    line = line.replace(" ", "-")
    line = line.replace("/", "-")
    line = line.replace("-&-", "-and-")
    line = line.replace("&", "-and-")
    line = unidecode(line).decode()
    line = urllib.parse.quote(line)
    line = f"https://genius.com/{line}-lyrics"

    if return_url:
        return line

    page = requests.get(line)
    html = BeautifulSoup(page.text, 'html.parser')

    result = html.find('div', class_='lyrics') #.get_text()
    if result is not None:
        lyrics = result.get_text()
        lyrics2 = []
        for line in lyrics.splitlines():
            if line.startswith("["):
                pass
            else:
                lyrics2.append(line)

        lyrics = "\n".join(lyrics2)
        lyrics = lyrics.strip("\n")
        return lyrics

    # New layout type
    else:

        results = html.findAll("div", {"class": lambda l: l and "Lyrics__Container" in l})
        lyrics = "".join([r.get_text("\n") for r in results])
        level = 0
        new = ""
        for cha in lyrics:
            if level <= 0:
                new += cha
            if cha == "[":
                level += 1
            if cha == "]":
                level -= 1
        lyrics = new

        lines = lyrics.splitlines()
        new_lines = []
        for line in lines:
            if "[" in line:
                line = line.split("[", 1)[0]
                if line:
                    line += "\n"

            new_lines.append(line.lstrip().rstrip(" ") + "\n")

        lyrics = "".join(new_lines)
        lyrics = lyrics.replace("(\n", "(")
        lyrics = lyrics.replace("\n)", ")")
        lyrics = lyrics.lstrip("\n")
        lyrics = lyrics.lstrip()
        return lyrics




lyric_sources = {
    # "Apiseeds": apiseeds,
    "Happi": happi,
    "Genius": genius,
    #"LyricWiki": lyricwiki,
    "lyrics.ovh": ovh,
}

uses_scraping = {
    #"LyricWiki",
    "Genius"
}
